import pandas as pd
import numpy as np
import datetime as dt
from tqdm import tqdm
import warnings

# Filter out the unnecessary warning message
warnings.filterwarnings("ignore", message="Converting to PeriodArray/Index representation will drop timezone information.", category=UserWarning)

def _preprocess_pipeline(log, start_date, end_date):
    ''' Executes the preprocessing pipeline that prepares the given log for plotting. 

        args:
            - log               : pd.DataFrame 
            - event_features    : list of strings, containing the column names of the event features to be preprocessed.
            - start_date        : string "dd/mm/YYYY"
            - end_date          : string "dd/mm/YYYY"
        
        returns:
            - log               : preprocessed pd.DataFrame
    
    '''

    # Verify whether for each case, the events (rows) are arranged in the ascending order of the timestamps.
    # If not, issue a warning. Order automatically fixed.
    log = order_events(log)
    
    # Filter time range if start_date and / or end_date given.
    if start_date or end_date:
        log = select_timerange(log, start_date = start_date, end_date = end_date)

    #Adding periodic timestamps
    log = add_periodic_timestamps(log)

    #Adding case durations (/throughput times):
    log = add_case_duration(log)

    # Adding num_events per case: 
    log = add_num_events(log)

    return log 

# Preprocessing functions called by the preprocessing pipeline:
def select_timerange(log, start_date, end_date):
    """Select only those cases starting after `start_date`, and ending 
    before `end_date`. 

    Parameters
    ----------
    log : pd.DataFrame
        The event log. 
    start_date : str
        Start date as a string. Format "dd/mm/YYYY".
    end_date : str
        End date as a string. Format "dd/mm/YYYY".

    Returns
    -------
    log : pd.DataFrame
        The event log filtered based on the given time range. 
    """

    if start_date:
        start_date = pd.to_datetime(start_date, format = "%d/%m/%Y").tz_localize('UTC')
    else: 
        start_date = log['time:timestamp'].min() #timestamp object 

    if end_date:
        end_date = pd.to_datetime(end_date, format = "%d/%m/%Y").tz_localize('UTC')
    else: 
        end_date = log['time:timestamp'].max() #timestamp object 
        end_date = end_date +  pd.DateOffset(months=1)

    # Get first and last timestamp for each unqiue case 
    first_last = log.pivot_table(values = 'time:timestamp', index = 'case:concept:name', aggfunc = ['min', 'max']).reset_index()
    first_last.columns = first_last.columns.get_level_values(0)
    first_last.columns = ['case:concept:name', '_first_stamp_case', '_last_stamp_case']
    
    # Temporarily add first and last timestamp to log 
    log = log.merge(first_last, on = 'case:concept:name', how = 'left')
    # Filter on case level
    log = log[(log['_first_stamp_case'] >= start_date) & (log['_last_stamp_case'] <= end_date)]
    # Drop temporary columns
    log = log.drop(columns=['_first_stamp_case', '_last_stamp_case'], axis = 1)
    
    return log

def order_events(log):
    ''' Verifies whether for each case, the events (rows) are arranged in the ascending order of the timestamps.
        If not, a warning is issued. The correct ordering is automatically done.
    '''
    local_log = log.copy().reset_index(drop = True)
    case_log = local_log.drop_duplicates(subset = 'case:concept:name').copy()
    case_log = case_log[['case:concept:name']]
    case_log.loc[:, 'int_id'] = [i for i in range(len(case_log))]
    local_log = local_log.merge(case_log, on = 'case:concept:name', how = 'left')
    local_log = local_log.sort_values(['int_id', 'time:timestamp'])
    original_sorted_index = pd.Series(local_log.index.copy())
    local_log.reset_index(drop = True, inplace = True)
    new_sorted_index = pd.Series(local_log.index)
    wrong_order = (new_sorted_index != original_sorted_index).any()

    if wrong_order:
        warning_message = get_warning_message()
        warnings.warn(warning_message)

    local_log = local_log.drop(['int_id'], axis = 1)
    return local_log


def add_periodic_timestamps(log):
    """Preprocesses the event log by adding periodic timestamps for each
    possible case grouping.

    The case grouping determines how each case is assigned to a certain 
    time period. This is jointly determined by both the 'frequency' and 
    'case_assignment' argument present in each plotting method of the 
    'DynamicLogPlots' class. A case feature column is added for each
    ('frequency', 'case_assignment') combination. 

    Parameters
    ----------
    log : pandas.DataFrame
        Current version of the event log. 

    Returns
    -------
    pandas.DataFrame
        The log enhanced with multiple columns. 
    """
    
    def add_two_weekly_timestamp():
        """Compute for each timestamp in the log the two-weekly timestamp
        that represents the beginning of the two-week period in which
        each event occurred. 

        Returns
        -------
        pandas.Series
            Series containing two-weekly timestamps for every event. 
        """
        weekly_timestamps_unique = list(set(list(log['weekly_timestamp'])))
        weekly_timestamps_unique.sort()
        num_weekly_stamps=len(weekly_timestamps_unique)
        ids_retained = [i for i in range(0, num_weekly_stamps, 2)]
        reduced_inds = []
        for ret_idx in ids_retained: 
            red_inds = [ret_idx for _ in range(2)]
            reduced_inds.extend(red_inds)
        reduced_inds= reduced_inds[:num_weekly_stamps]
        reduced_timestamps= []
        for idx in reduced_inds:
            reduced_timestamps.append(weekly_timestamps_unique[idx])
        reduce_dict= dict(zip(weekly_timestamps_unique, reduced_timestamps))
        all_weekly_timestamps = list(log['weekly_timestamp'])
        reduced_timestamps_all = []
        for wkly_stamp in tqdm(all_weekly_timestamps):
            red_stamp = reduce_dict[wkly_stamp]
            reduced_timestamps_all.append(red_stamp)
        two_weekly_timestamp = reduced_timestamps_all
        return two_weekly_timestamp
    def add_halfyearly_timestamp():
        """Compute for each timestamp in the log the half-yearly timestamp
        that represents the beginning of the 6-month period in which
        each event occurred. 

        Returns
        -------
        pandas.Series
            Series containing half_yearly timestamps for every event.
        """

        # Series of integers referring to the month
        month_integer = timestamps.dt.month

        date_minus_3_months = timestamps - pd.DateOffset(months=3)
        minus_3_quarterly_timestamp = date_minus_3_months.dt.to_period("Q").astype('datetime64[ns]')

        second_or_fourth_quarter = (month_integer.isin([4,5,6])) | (month_integer.isin([10,11,12]))

        # Computing half-yearly timestamp. 
        half_yearly_timestamp = np.where(second_or_fourth_quarter, minus_3_quarterly_timestamp, 
                                                log['quarterly_timestamp'])
        return half_yearly_timestamp
    
    def add_first_timestamp(log):
        """Add for every periodic timestamp an additional case feature
        column that indicates the period in which its first event 
        occurred. 

        Parameters
        ----------
        log : pandas.DataFrame
            Current version of the event log. 

        Returns
        -------
        pandas.DataFrame
            Current log object with 'len(periodic_timestamps)' additional cols. 
        """
        extended_PerTS = ['case:concept:name'] + periodic_timestamps
        local_log = log.drop_duplicates(subset = 'case:concept:name')[extended_PerTS].copy()
        new_cols = ['case:concept:name'] + [col + '_firstev' for col in periodic_timestamps]
        local_log.columns = new_cols
        log = log.merge(local_log, on = 'case:concept:name', how = 'left')
        return log 

    def add_last_timestamp(log):
        """Add for every periodic timestamp an additional case feature
        column that indicates the period in which its last event 
        occurred. 

        Parameters
        ----------
        log : pandas.DataFrame
            Current version of the event log. 

        Returns
        -------
        pandas.DataFrame
            Current log object with 'len(periodic_timestamps)' additional cols. 
        """
        extended_PerTS = ['case:concept:name'] + periodic_timestamps
        local_log = log.groupby('case:concept:name', sort= False).last().reset_index()[extended_PerTS].copy()
        new_cols = ['case:concept:name'] + [col + '_lastev' for col in periodic_timestamps]
        local_log.columns = new_cols
        log = log.merge(local_log, on = 'case:concept:name', how = 'left')
        return log 

    def add_maxevents_timestamp(log, time_col):
        """ Add for one periodic timestamp 'time_col' in periodic_timestamps
        an additional case feature (for each case) that stores the 
        'time_col' value (i.e. period) in which most of its events occurred. 

        Parameters
        ----------
        log : pandas.DataFrame
            Current version of the event log. 
        time_col : str
            Indicates the added periodic timestamp column. 

        Returns
        -------
        pd.DataFrame
            Current log object with one additional column. 
        """
        local_log = log[['case:concept:name', 'concept:name', time_col]].copy()
        # For each case: add number of events per period of time_col
        local_log['per_numevs'] = local_log.groupby(['case:concept:name', time_col])['concept:name'].transform('count')
        # For each case: add max 'per_numevs' to each row corresponding to that case
        local_log['max_pernumevs'] = local_log.groupby('case:concept:name')['per_numevs'].transform('max')
        # For each case: filter out the period (time_col stamp) in which most of its events were executed. Ties are broken by taking the first timestamp:
        local_log = local_log[local_log['per_numevs'] == local_log['max_pernumevs']].drop_duplicates(subset= 'case:concept:name')[['case:concept:name', time_col]].copy()
        local_log.columns = ['case:concept:name', time_col+'_maxev']
        # Merging the resulting df back into the original log: 
        log = log.merge(local_log, on = 'case:concept:name', how= 'left')
        return log 
    

    timestamps = log['time:timestamp'].copy()
    # 1-minute interval
    log['1min_timestamp'] = timestamps.dt.floor("1min")

    # 5-minute interval
    log['5min_timestamp'] = timestamps.dt.floor("5min")

    # 10-minute interval
    log['10min_timestamp'] = timestamps.dt.floor("10min")

    # 30-minute interval
    log['30min_timestamp'] = timestamps.dt.floor("30min")

    # 1h interval
    log['hourly_timestamp'] = timestamps.dt.floor("1H")

    # 2-hour interval
    log['2hour_timestamp'] = timestamps.dt.floor("2H")

    # 12-hour interval
    log['12hour_timestamp'] = timestamps.dt.floor("12H")

    # Daily interval:
    log['daily_timestamp'] = timestamps.dt.floor("D")

    # Weekly interval:
    log['weekly_timestamp'] = timestamps.dt.to_period("W").astype('datetime64[ns]')

    # also adding a two_weekly_timestamp: 
    two_weekly_timestamp = add_two_weekly_timestamp()
    log['two_weekly_timestamp'] = two_weekly_timestamp

    # monthly interval
    log['monthly_timestamp'] = timestamps.dt.to_period("M").astype('datetime64[ns]')

    # quarterly interval
    log['quarterly_timestamp'] = timestamps.dt.to_period("Q").astype('datetime64[ns]')

    half_yearly_timestamp = add_halfyearly_timestamp()
    log['half_yearly_timestamp'] = half_yearly_timestamp

    # Add 3 different types of periodic case assignment timestamps: 
    periodic_timestamps = ['1min_timestamp', '5min_timestamp', '10min_timestamp', 
                           '30min_timestamp', 'hourly_timestamp', '2hour_timestamp', 
                           '12hour_timestamp', 'daily_timestamp', 'weekly_timestamp', 
                           'two_weekly_timestamp', 'monthly_timestamp', 
                           'quarterly_timestamp', 'half_yearly_timestamp']
    
    log = add_first_timestamp(log)
    log = add_last_timestamp(log)

    for time_col in periodic_timestamps:
        log = add_maxevents_timestamp(log, time_col)
    
    # Drop the former periodic timestamp columns, since they are not needed anymore: 
    log = log.drop(periodic_timestamps, axis = 1)

    return log 
    
def add_case_duration(log):
    '''
        args: 
            - log:  complete pd.DataFrame 
        returns:
            - log:  pd.DataFrame enhanced with throughput time columns that indicate the case duration. The added columns are:
                    'tt_microseconds', 'tt_milliseconds', 'tt_seconds', 'tt_minutes', 'tt_hours', 'tt_days' and 'tt_weeks'
        
    '''
    local_log = log[['case:concept:name', 'time:timestamp']].copy()
    times = local_log.groupby('case:concept:name', sort= False)['time:timestamp'].agg(['min', 'max']).reset_index()
    times['delta'] = times['max'].copy() - times['min'].copy()
    times['tt_microseconds']  = times['delta'].copy() / pd.Timedelta(microseconds = 1)
    times['tt_milliseconds']  = times['delta'].copy() / pd.Timedelta(milliseconds = 1)
    times['tt_seconds']       = times['delta'].copy() / pd.Timedelta(seconds = 1)
    times['tt_minutes']       = times['delta'].copy() / pd.Timedelta(minutes = 1)
    times['tt_hours']         = times['delta'].copy() / pd.Timedelta(hours = 1)
    times['tt_days']          = times['delta'].copy() / pd.Timedelta(days = 1)
    times['tt_weeks']         = times['delta'].copy() / pd.Timedelta(weeks = 1)
    times = times[['case:concept:name', 'tt_microseconds', 'tt_milliseconds', 'tt_seconds', 'tt_minutes', 'tt_hours', 'tt_days', 'tt_weeks']].copy()
    log = log.merge(times, on = 'case:concept:name', how = 'left')
    return log


def add_num_events(log):
    '''
        args: 
            - log: complete pd.DataFrame 
        returns:
            - log: pd.DataFrame enhanced with the 'num_events' column, indicating for each case how many events a case has. 
    '''
    events_case= log.pivot_table(values='concept:name', index= 'case:concept:name', aggfunc= 'count', fill_value=0).reset_index()
    events_case.columns = ['case:concept:name', 'num_events']
    log = log.merge(events_case, how='left', on='case:concept:name')
    return log 


def get_warning_message():
    warning_message = "In some cases in the given log, the events were not ordered correctly based on their timestamp. \
In the DynamicLogPlots instance' internal representation of the log, this problem is resolved by correctly sorting the events."
    return warning_message


import DyLoPro.check_args as ca
import pandas as pd
import numpy as np

# Some checks are identical for all plotting functions, these are grouped in the _verify_common() method:
def _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem):
    """Verify validity of the hyperparameters common to all plotting methods. 

    Parameters
    ----------
    time_unit : str
        Time unit in which the throughput time of cases is specified. 
        The options are {'microseconds', 'milliseconds', 'seconds', 'minutes', 
        'hours', 'days', 'weeks'}. 
    frequency : str
        Frequency by which the observations are grouped together. The options are
        {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
        '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
        Frequency by which the observations are grouped together, by default 'weekly'.
    case_assignment : str
        Condition upon which a case is assigned to a certain period. 
        Options are {'first_event', 'last_event', 'max_events'}.
    numeric_agg : str
        _description_
    xtr_outlier_rem : bool
        _description_

    Returns
    -------
    _type_
        _description_
    """
    
    ''' Verification procedure of the arguments passed to this function is identical for all plotting methods.
        Therefore, these procedures are grouped as one in this function. 

        returns: 
            - error     : bool; True if an argument is specified incorrectly, False otherwise. 
            - err_msge  : string; contains the error-specific error message if any.
    '''
    error, err_msge = False, " "

    # verify time_unit:
    if time_unit != 'days':
        error, err_msge = ca._verify_time_unit(time_unit)
        if error:
            return error, err_msge
    
    # verify frequency 
    if frequency != 'weekly':
        error, err_msge = ca._verify_frequency(frequency)
        if error:
            return error, err_msge
    
    # verify case_assignment
    if case_assignment != 'first_event':
        error, err_msge = ca._verify_case_assignment(case_assignment)
        if error:
            return error, err_msge
    
    # verify numeric_agg
    error, err_msge = ca._verify_numeric_agg(numeric_agg)
    if error:
        return error, err_msge
    
    # verify xtr_outlier_rem:
    error, err_msge = ca._verify_xtr_outlier_rem(xtr_outlier_rem)
    if error:
        return error, err_msge
    
    return error, err_msge



def _verify_topKdfr(outcome, time_unit, frequency, case_assignment, plt_type, numeric_agg, max_k, xtr_outlier_rem):
    '''
    Checks whether all arguments are correctly specified for the topK_dfr_evol() method.

    '''

    if plt_type != 'univariate':
        error, err_msge = ca._verify_typedfr(plt_type)
        if error:
            raise ValueError(err_msge)

    # verify outcome (if plt_type = 'type_outcome):
    if plt_type == 'type_outcome':
        error, err_msge = ca._verify_outcomeCol(outcome)
        if error:
            raise ValueError(err_msge)

    # common verifications:
    error, err_msge = _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem)
    if error:
        raise ValueError(err_msge)
    
    # verify max_k: 
    error, err_msge = ca._verify_max_k(max_k)
    if error: 
        raise ValueError(err_msge)



def _verify_dfr(dfrelations, directly_follows_relations, outcome, time_unit, frequency, case_assignment, plt_type, numeric_agg, xtr_outlier_rem):
    '''
    Checks whether all arguments are correctly specified for the dfr_evol() method.

    '''

    # verify plt_type: 
    if plt_type != 'univariate':
        error, err_msge = ca._verify_typedfr(plt_type)
        if error:
            raise ValueError(err_msge)

    # verify outcome (if plt_type = 'type_outcome):
    if plt_type == 'type_outcome':
        error, err_msge = ca._verify_outcomeCol(outcome)
        if error:
            raise ValueError(err_msge)

    # common verifications:
    error, err_msge = _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem)
    if error:
        raise ValueError(err_msge)

    # Verify the passed on dfr's in the directly_follows_relations list:
    error, err_msge = ca._verify_dfrelations(dfrelations, directly_follows_relations)
    if error:
        raise ValueError(err_msge)


def _verify_topKvariants(outcome, time_unit, frequency, case_assignment, plt_type, numeric_agg, max_k, xtr_outlier_rem):
    '''
    Checks whether all arguments are correctly specified for the topK_variants_evol() plotting method.
    '''
    
    # verify plt_type: 
    if plt_type != 'univariate':
        error, err_msge = ca._verify_typevar(plt_type)
        if error:
            raise ValueError(err_msge)

    # verify outcome (if plt_type = 'type_outcome):
    if plt_type == 'type_outcome':
        error, err_msge = ca._verify_outcomeCol(outcome)
        if error:
            raise ValueError(err_msge)

    # common verifications:
    error, err_msge = _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem)
    if error:
        raise ValueError(err_msge)
    
    # verify max_k: 
    error, err_msge = ca._verify_max_k(max_k)
    if error: 
        raise ValueError(err_msge)

def _verify_variants(all_vars, variants, outcome, time_unit, frequency, case_assignment, plt_type, numeric_agg, xtr_outlier_rem):
    '''
    Checks whether all arguments are correctly specified for the variants_evol() plotting method.
    '''

    # verify plt_type: 
    if plt_type != 'univariate':
        error, err_msge = ca._verify_typevar(plt_type)
        if error:
            raise ValueError(err_msge)

    # verify outcome (if plt_type = 'type_outcome):
    if plt_type == 'type_outcome':
        error, err_msge = ca._verify_outcomeCol(outcome)
        if error:
            raise ValueError(err_msge)

    # common verifications:
    error, err_msge = _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem)
    if error:
        raise ValueError(err_msge)

    # Verify the passed on variants in the 'variants' list:
    error, err_msge = ca._verify_variantslist(all_vars, variants)
    if error:
        raise ValueError(err_msge)

def _verify_topKcatFt(feature, ftr_type, feature_list, outcome, time_unit, frequency, case_assignment, plt_type, numeric_agg, max_k, xtr_outlier_rem):
    '''
    Checks whether all arguments are correctly specified for the topK_categorical_caseftr_evol() or for the topK_categorical_eventftr_evol() plotting methods.

    args:
        - feature       : string; column name of categorical feature to be plotted
        - ftr_type      : string = 'case' or 'event'; specifying whether the categorical feature is a case or an event feature. 
        - feature_list  : list of strings; list of the categorical case or event features names upon initialization of the DynamicLogPlots instance. 
    '''

    # verify plt_type: 
    if plt_type != 'univariate':
        error, err_msge = ca._verify_typeftrs(plt_type)
        if error:
            raise ValueError(err_msge)

    # verify outcome (if plt_type = 'type_outcome):
    if plt_type == 'type_outcome':
        error, err_msge = ca._verify_outcomeCol(outcome)
        if error:
            raise ValueError(err_msge)

    # common verifications:
    error, err_msge = _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem)
    if error:
        raise ValueError(err_msge)
    
    # verify max_k: 
    error, err_msge = ca._verify_max_k(max_k)
    if error: 
        raise ValueError(err_msge)
    
    # verify whether feature already in appropriate feature_list
    if feature not in feature_list:
        err_msge = "Column '{}' not yet added to the categorical_{}features list. Add it by means of \
calling the add_categorical_{}ft({}) on your DynamicLogPlots instance.".format(feature, ftr_type, ftr_type, feature)
        raise ValueError(err_msge)

def _verify_numFt(features, ftr_type, feature_list, outcome, time_unit, frequency, case_assignment, 
                  plt_type, numeric_agg, xtr_outlier_rem, numEventFt_transform = None):
    '''
    Checks whether all arguments are correctly specified for the num_casefts_evol() or for the num_eventfts_evol() plotting methods.

    args:
        - features      : list of strings; containing the column names of numerical features to be plotted
        - ftr_type      : string = 'case' or 'event'; specifying whether the numerical features are case or event features. 
        - feature_list  : list of strings; list of the numerical case or event features names upon initialization of the DynamicLogPlots instance. 
    '''
    if type(features) != list:
        err_msge = "The numerical features that need to be plotted should be passed on as a list of strings."
        raise TypeError(err_msge)

    # verify plt_type: 
    if plt_type != 'univariate':
        error, err_msge = ca._verify_typeftrs(plt_type)
        if error:
            raise ValueError(err_msge)

    # verify outcome (if plt_type = 'type_outcome):
    if plt_type == 'type_outcome':
        error, err_msge = ca._verify_outcomeCol(outcome)
        if error:
            raise ValueError(err_msge)

    # common verifications:
    error, err_msge = _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem)
    if error:
        raise ValueError(err_msge)
    
    # verify whether the features already contained within the appropriate feature_list
    for feature in features:
        if feature not in feature_list:
            err_msge = "Column '{}' not yet added to the numerical_{}features list. Add it by means of \
calling the add_numerical_{}ft({}) on your DynamicLogPlots instance.".format(feature, ftr_type, ftr_type, feature)
            raise ValueError(err_msge)
    
    # in case of numerical event features, an additional parameter needs to be checked: 
    if ftr_type == 'event':
        error, err_msge = ca._verify_numEventFt_transform(numEventFt_transform)
        if error:
            raise ValueError(err_msge)

def _verify_distinctvars(outcome, time_unit, frequency, case_assignment, plt_type, numeric_agg, xtr_outlier_rem, cases_initialized):
    '''
    Checks whether all arguments are correctly specified for the distinct_variants_evol() and distinct_variants_AdvancedEvol() plotting methods.
    '''
    # verify plt_type: 
    if plt_type != 'univariate':
        error, err_msge = ca._verify_typeDistVars(plt_type)
        if error:
            raise ValueError(err_msge)

    # verify outcome (if plt_type = 'type_outcome):
    if plt_type == 'type_outcome':
        error, err_msge = ca._verify_outcomeCol(outcome)
        if error:
            raise ValueError(err_msge)

    # common verifications:
    error, err_msge = _verify_common(time_unit, frequency, case_assignment, numeric_agg, xtr_outlier_rem)
    if error:
        raise ValueError(err_msge)
    
    # verify cases_initialized:
    error, err_msge = ca._verify_cases_initialized(cases_initialized = cases_initialized)
    if error:
        raise TypeError(err_msge)


def _verify_initial(log, case_id_key, activity_key, timestamp_key, categorical_casefeatures, 
                    numerical_casefeatures, categorical_eventfeatures, numerical_eventfeatures, outcome, start_date, end_date):
    
    # Verify whether the log is actually a pd.DataFrame:
    print("Verification step 1/10", end = '\r')
    if type(log) != pd.core.frame.DataFrame:
        raise TypeError("The log object should be a pandas DataFrame. Convert the log to a pandas DataFrame first.")

    # Verify whether there are no missing values for the case_id_key: 
    print("Verification step 2/10", end = '\r')
    if log[case_id_key].isna().any():
        err_msge = "Column '{}', which should contain the case IDs, contains missing values. Please make sure that every row contains a valid case id."

    # Verify whether the given columns are actually in the passed on log
    print("Verification step 3/10", end = '\r')
    cols = list(log.columns)
    given_cols = [case_id_key, activity_key, timestamp_key] # + categorical_casefeatures + numerical_casefeatures + categorical_eventfeatures + numerical_eventfeatures
    error, err_msge = ca._verify_ColsInLog(cols, given_cols)
    if error:
        raise KeyError(err_msge)
    

    # Verify whether the timestamp_key column is actually a datetime64 column:
    print("Verification step 4/10", end = '\r')
    if not pd.api.types.is_datetime64_any_dtype(log[timestamp_key]):
        raise TypeError("Column '{}', which should contain the timestamps, is not of a datetime64 dtype.".format(timestamp_key))

    # Verify the numerical case features
    print("Verification step 5/10", end = '\r')
    for num_ft in numerical_casefeatures:
        error, err_msge, err_cat = ca._verify_newNumCaFt(log, case_feature = num_ft, case_id_key = case_id_key)
        if error:
            if err_cat == 'type_error':
                raise TypeError(err_msge)
            elif err_cat == 'key_error':
                raise KeyError(err_msge)
            else:
                raise ValueError(err_msge)
    print("Verification step 6/10", end = '\r')
    # Verify the numerical event features:
    for num_ft in numerical_eventfeatures:
        error, err_msge, err_cat = ca._verify_newNumEvFt(log, event_feature = num_ft)
        if error:
            if err_cat == 'type_error':
                raise TypeError(err_msge)
            elif err_cat == 'key_error':
                raise KeyError(err_msge)
            else:
                raise ValueError(err_msge)
    
    # Verify the categorical case features:
    print("Verification step 7/10", end = '\r')
    for cat_ft in categorical_casefeatures:
        error, err_msge, err_cat = ca._verify_newCatCaFt(log, case_feature = cat_ft, case_id_key = case_id_key)
        if error:
            if err_cat == 'key_error':
                raise KeyError(err_msge)
            elif err_cat == 'type_error':
                raise TypeError(err_msge)
            else:
                raise ValueError(err_msge)

    # Verify the categorical event features:
    print("Verification step 8/10", end = '\r')
    for cat_ft in categorical_eventfeatures:
        error, err_msge, err_cat = ca._verify_newCatEvFt(log, event_feature = cat_ft)
        if error:
            if err_cat == 'key_error':
                raise KeyError(err_msge)
            elif err_cat == 'type_error':
                raise TypeError(err_msge)
            else:
                raise ValueError(err_msge)

    # Verify start and / or end date if specified 
    print("Verification step 9/10", end = '\r')
    if start_date or end_date: 
        error, err_msge, err_cat = ca._verify_start_end_date(start_date, end_date)
        if error:
            if err_cat == 'type_error':
                raise TypeError(err_msge)
            else:
                raise ValueError(err_msge)
                
    # Verify outcomn column if specified. 
    print("Verification step 10/10", end = '\r')
    if outcome: #If not None
        error, err_msge = ca._verify_ColsInLog(cols, [outcome])
        if error:
            raise KeyError(err_msge)
        error, err_msge = ca._verify_outcomeCol_initial(log, outcome, case_id_key = case_id_key)
        if error: 
            raise ValueError(err_msge)


# Utility method specific checks 

#   Utility method to verify the to-be added categorical case feature 
def _verify_addCatCaFt(log, case_feature):
    ''' Validity check specifically for verifying whether the to-be added case_feature in the add_categorical_caseft(self, case_feature)
        method is valid. Raises a ValueError if not valid. 

        args:
            - log: pd.DataFrame 
            - case_feature: string containing the column name (in log) of the categorical case feature to-be added. 
    '''
    error, err_msge, err_cat = ca._verify_newCatCaFt(log, case_feature)
    if error:
        if err_cat == 'key_error':
            raise KeyError(err_msge)
        elif err_cat == 'type_error':
            raise TypeError(err_msge)
        else:
            raise ValueError(err_msge)


#   Utility method to verify the to-be added numerical case feature 
def _verify_addNumCaFt(log, case_feature):
    ''' Validity check specifically for verifying whether the to-be added case_feature in the add_numerical_caseft(self, case_feature)
        method is valid. Raises a ValueError if not valid. 

        args:
            - log: pd.DataFrame 
            - case_feature: string containing the column name (in log) of the numerical case feature to-be added. 
    '''
    error, err_msge, err_cat = ca._verify_newNumCaFt(log, case_feature)
    if error:
        if err_cat == 'key_error':
            raise KeyError(err_msge)
        elif err_cat == 'type_error':
            raise TypeError(err_msge)
        else:
            raise ValueError(err_msge)
    
#   Utility method to verify the to-be added categorical event feature 
def _verify_addCatEvFt(log, event_feature):
    ''' Validity check specifically for verifying whether the to-be added event_feature in the add_categorical_eventft(self, event_feature)
        method is valid. Raises a ValueError if not valid. 

        args:
            - log: pd.DataFrame 
            - case_feature: string containing the column name (in log) of the categorical case feature to-be added. 
    '''
    error, err_msge, err_cat = ca._verify_newCatEvFt(log, event_feature)
    if error:
        if err_cat == 'key_error':
            raise KeyError(err_msge)
        else:
            raise TypeError(err_msge)

#   Utility method to verify the to-be added numerical event feature 
def _verify_addNumEvFt(log, event_feature):
    ''' Validity check specifically for verifying whether the to-be added event_feature in the add_numerical_eventft(self, event_feature)
        method is valid. Raises a ValueError if not valid. 

        args:
            - log: pd.DataFrame 
            - event_feature: string containing the column name (in log) of the numerical event feature to-be added. 
    '''
    error, err_msge, err_cat  = ca._verify_newNumEvFt(log, event_feature)
    if error:
        if err_cat == 'key_error':
            raise KeyError(err_msge)
        else:
            raise TypeError(err_msge)

#   Utility method to verify the to-be added outcome column
def _verify_addOutCol(log, outcome):
    ''' Validity check specifically for verifying whether the to-be added event_feature in the add_numerical_eventft(self, event_feature)
        method is valid. Raises a ValueError if not valid. 

        args:
            - log: pd.DataFrame 
            - outcome: string containing the column name (in log) of the outcome feature to be added. 
    '''

    cols = list(log.columns)
    error, err_msge = ca._verify_ColsInLog(cols, [outcome])
    if error:
        raise KeyError(err_msge)
    error, err_msge = ca._verify_outcomeCol_initial(log, outcome, case_id_key = 'case:concept:name')
    if error: 
        raise ValueError(err_msge)


def _verify_select_time_range(start_date, end_date):
    '''Validity check for verifyign the start_date and end_date specification, if any. Specificaly for the 
    select_time_range(start_date, end_date) method. 
    '''

    if start_date or end_date: 
        error, err_msge, err_cat = ca._verify_start_end_date(start_date, end_date)
        if error:
            if err_cat == 'type_error':
                raise TypeError(err_msge)
            else:
                raise ValueError(err_msge)

def _verify_get_DFR_df(max_k, directly_follows_relations, counts, dfrelations):
    """Verify the input parameters of the `get_DFR_df()` method 
    of the `DynamicLogPlots` class. 

    Parameters
    ----------
    max_k : int
    directly_follows_relations : list of tuple
        List of DFRs specified by the user. 
    counts : bool 
        Boolean indicating whether DFR counts should 
        be included. 
    dfrelations: list of tuple
        List of all DFRs present in the event log. 
    """
    # Verify counts:
    error, err_msge = ca._verify_counts_bool(counts)
    if error: 
        raise TypeError(err_msge)

    # Verify max_k 
    if max_k:
        error, err_msge = ca._verify_max_k(max_k)
        if error:
            raise ValueError(err_msge)
    # If max_k = None, directly_follows_relations should be (correctly)
    # specified. 
    else:
        # Check whether directly_follows_relations not None: 
        error, err_msge = ca._verify_dfrlist_None(directly_follows_relations)
        if error: 
            raise ValueError(err_msge)
        # Check whether DFRs exist and / or correctly specified:
        error, err_msge = ca._verify_dfrelations(dfrelations, directly_follows_relations)
        if error: 
            raise ValueError(err_msge)

def _verify_get_var_df(max_k, variants, counts, all_vars):
    """Verify the input parameters of the `get_var_df()` method 
    of the `DynamicLogPlots` class. 

    Parameters
    ----------
    max_k : int
    variants : list of tuple
        List of variants specified by the user. 
    counts : bool 
        Boolean indicating whether DFR counts should 
        be included. 
    all_vars: list of tuple
        List of all variants present in the event log. 
    """ 
    # Verify counts:
    error, err_msge = ca._verify_counts_bool(counts)
    if error: 
        raise TypeError(err_msge)

    # Verify max_k 
    if max_k:
        error, err_msge = ca._verify_max_k(max_k)
        if error:
            raise ValueError(err_msge)
    # If max_k = None, directly_follows_relations should be (correctly)
    # specified. 
    else:
        # Check whether directly_follows_relations not None: 
        error, err_msge = ca._verify_varlist_None(variants)
        if error: 
            raise ValueError(err_msge)
        # Check whether the variants exist and / or correctly specified:
        error, err_msge = ca._verify_variantslist(all_vars, variants)
        if error:
            raise ValueError(err_msge)
    


""" General function called upon by every plotting function. """
def plt_period(x, y, axes, y_label, number=None, max_k=None, title=None, label=None,  location=None, color=None):
    """Generic plotting function. 4 options:

    #. Data needs to be plotted on a figure with 1 other plot, with each of 
       the 2 plots a different vertical axis (left and right). This data is 
       assigned to the left vertical axis. Needed arguments: the required 
       arguments, label = string, location = 'left', color = string.
    
    #. Data needs to be plotted on a figure with 1 other plot, with each of 
       the 2 plots a different vertical axis (left and right). This data is 
       assigned to the right vertical axis. Needed arguments: the required 
       arguments, label = string, location = 'right', color = string.
    
    #. Data needs to be plotted on a figure that will not contain any other 
       plots. Needed arguments: the required arguments, title = string.
    
    #. Data needs to be plotted on a figure that will contain several other 
       plots, but all of the same quantity / unit. Hence, only one vertical 
       axis (the left one) will be used by max_k plots. Needed arguments: 
       the required arguments, number = int, max_k = int, title = string, 
       label = string.

    Parameters
    ----------
    x : _type_
        _description_
    y : _type_
        _description_
    axes : _type_
        _description_
    y_label : _type_
        _description_
    number : _type_, optional
        _description_, by default None
    max_k : _type_, optional
        _description_, by default None
    title : _type_, optional
        _description_, by default None
    label : _type_, optional
        _description_, by default None
    location : _type_, optional
        _description_, by default None
    color : _type_, optional
        _description_, by default None
    """
    marker_list = ['o', 's', '+', 'x', 'D', 'P', '1', 'v', '*', '>', '<']
    num_markers = len(marker_list)
    
    if color:
        if location=='left': #1
            axes.plot(x, y, label=label, color = color, marker='o')
            axes.legend(loc='upper left')
            axes.grid(True, axis='x')
            axes.set_ylabel(y_label, color=color)
        elif location=='right': #2
            # axes.plot(x, y, label="DFR {}".format(number), color = color, marker='o')
            axes.plot(x, y, label=label, color = color, marker='s')
            axes.legend(loc='upper right')
            # axes.set_ylabel("Number of DFR {} per case".format(number), color=color)
            axes.set_ylabel(y_label, color=color)
        if title:
            axes.set_title(title)

    else:
        if label: #4
            # Marker_idx configured st in case of number > num_markers, 
            # we start again at the beginning of the marker list. 
            marker_idx = number - 1 - num_markers*((number-1)//num_markers)
            axes.plot(x, y, label=label, marker= marker_list[marker_idx])
            if number==max_k:
                if max_k > 2: 
                    axes.legend(loc="upper left", fontsize='x-small')
                else:
                    axes.legend()
                # axes.set_ylabel("# Directly-Follows Relations (DFR) per Case")
                axes.set_ylabel(y_label)
                axes.grid(True)
                # axes.set_title("{} evolution of #occurances/case for the {} most common Direclty-Follows Relationships (DFRs)".format(frequency, max_k))
                axes.set_title(title)
        else: #3
            axes.plot(x, y, marker='o')
            # axes.set_ylabel("Number of DFR {} per case".format(number))
            axes.set_ylabel(y_label)
            axes.grid(True)
            # axes.set_title("DFR {}: {} evolution of #occurances/case".format(number, frequency))
            axes.set_title(title)


"""Top-level package for DyLoPro.
DyLoPro is a generic visual analytics framework that 
facilitates the comprehensive exploration of the dynamics 
in event logs over time. It caters to all kinds of event logs, 
and provides the means to explore the dynamics from all 
different perspectives, both individually as well
as combined with the performance perspective. 
Consequently, also a multi-dimensional view on the 
dynamics in process performance can be adopted. The 
purpose of DyLoPro is to efficiently identify patterns, 
temporary or permanent changes and trends of interest
from different perspectives, which upon identification 
could be further analyzed in a manner deemed appropriate 
by the user.
"""
__author__ = """Brecht Wuyts """
__email__ = 'brecht.wuyts@kuleuven.be'
__version__ = '0.1.1'

from DyLoPro.DynamicLogPlotting import DynamicLogPlots


from DyLoPro.preprocess_utils import _preprocess_pipeline, select_timerange
import DyLoPro.univariate_plots as up
from DyLoPro.plotting_utils import get_variant_case, get_sorted_DFRs, get_filtered_dfr_df, get_filtered_var_df
import DyLoPro.validate_methods as valm
import pandas as pd

    

class DynamicLogPlots():
    def __init__(self, event_log, case_id_key = 'case:concept:name', activity_key = 'concept:name', timestamp_key = 'time:timestamp',
                    categorical_casefeatures = [], numerical_casefeatures = [], categorical_eventfeatures = [], numerical_eventfeatures = [], 
                    start_date = None, end_date = None, outcome = None):
        """Initialize a `DynamicLogPlots` instance by specifying the appropriate arguments.

        After successfully initializing a `DynamicLogPlots` object, all of DyLoPro's 
        plotting functionalities can invoked by calling the appropriate 
        `DynamicLogPlots` methods. 

        Parameters
        ----------
        event_log : pandas.DataFrame
            Event log. Events are regarded as instantaneous.  
        case_id_key : str, optional
            Column name (in `event_log`) containing the case ID. All 
            events pertaining to the same case should share the same 
            unique case ID. By default `'case:concept:name'`. 
        activity_key : str, optional
            Column name (in `event_log`) containing the activity labels. 
            By default `'concept:name'`. 
        timestamp_key : str, optional
            Column name (in `event_log`) containing the timestamps for 
            each event. By default `'time:timestamp'`. Should be of a 
            datetime64 dtype. 
        categorical_casefeatures : list of str, optional
            List of strings containing the column names (in `event_log`) 
            that correspond to categorical case features. All categorical 
            case features for which you wish to analyze the dynamics over 
            time must be specified in this list first, or alternatively, 
            after having already initialized a `DynamicLogPlots` object, 
            with the `add_categorical_caseft(case_feature)` class method. 
            By default `[]`. See Notes for more details on how each 
            categorical case feature column should be formatted. 
        numerical_casefeatures : list of str, optional
            List of strings containing the column names (in `event_log`) 
            that correspond to numerical case features. All numerical 
            case features for which you wish to analyze the dynamics over 
            time must be specified in this list first, or alternatively, 
            after having already initialized a `DynamicLogPlots` object, 
            with the `add_numerical_caseft(case_feature)` class method. 
            By default `[]`. See Notes for more details on how each 
            numerical case feature column should be formatted. 
        categorical_eventfeatures : list of str, optional
            List of strings containing the column names (in `event_log`) 
            that correspond to categorical event features. All 
            categorical event features for which you wish to analyze the 
            dynamics over time must be specified in this list first, or 
            alternatively, after having already initialized a 
            `DynamicLogPlots` object, with the 
            `add_categorical_eventft(event_feature)` class method. By 
            default `[]`. See Notes for more details on how each 
            categorical event feature column should be formatted.
        numerical_eventfeatures : list of str, optional
            List of strings containing the column names (in `event_log`) 
            that correspond to numerical event features. All numerical 
            event features for which you wish to analyze the dynamics 
            over time must be specified in this list first, or 
            alternatively, after having already initialized a 
            `DynamicLogPlots` object, with the 
            `add_numerical_eventft(event_feature)` class method. By 
            default `[]`. See Notes for more details on how each 
            numerical event feature column should be formatted.
        start_date : str, optional
            By default `None`. If specified, only the cases starting 
            after that date will be included in the dynamic profiling. 
            Should be specified in the following format `'dd/mm/YYYY'`. 
            For example, November 4th 2022 should be specified as 
            `start_date='04/11/2022'`.
        end_date : str, optional
            By default `None`. If specified, only the cases ending 
            before that date will be included in the dynamic profiling. 
            Should be specified in the following format `'dd/mm/YYYY'`. 
            For example, November 4th 2022 should be specified as 
            `start_date='04/11/2022'`.
        outcome : str, optional
            Column name (in `event_log`) containing the binary case 
            outcome values (if present). By default `None`. Should be of 
            an integer dtype. See Notes for more details on how an 
            outcome column should be formatted.

        Notes
        -----
        Formatting requirements specified columns in `event_log`:

        * `categorical_casefeatures` : Every column in `event_log` 
          specified in this list has to be of one of the following dtypes: 
          category, object, boolean. Furthermore, every event (row) 
          pertaining to the same case (i.e. same case ID specified in the 
          `case_id_key` column) should share the exact same value for each 
          case feature.

        * `numerical_casefeatures` : Every column in `event_log` 
          specified in this list has to be of a numerical dtype. 
          Furthermore, every event (row) pertaining to the same case (i.e. 
          same case ID specified in the `case_id_key` column) should share 
          the exact same value for each case feature.

        * `categorical_eventfeatures` : Every column in `event_log` 
          specified in this list has to be of one of the following dtypes: 
          category, object, boolean.

        * `numerical_eventfeatures` : Every column in `event_log` 
          specified in this list has to be of a numerical dtype.

        * `outcome`: An outcome column in `event_log` should be of an 
          integer dtype, only contain the values `1` (positive cases) and 
          `0` (negative cases). We regard outcome as case outcomes, and 
          hence every event (row) pertaining to the same case (i.e. 
          same case ID specified in the `case_id_key` column) should share 
          the exact same value for the outcome.

        """
        log = event_log.copy()
        valm._verify_initial(log, case_id_key, activity_key, timestamp_key, categorical_casefeatures, 
                    numerical_casefeatures, categorical_eventfeatures, numerical_eventfeatures, outcome, start_date, end_date)

        if case_id_key != 'case:concept:name':
            log.rename(columns = {case_id_key: 'case:concept:name'}, inplace = True)
        if activity_key != 'concept:name':
            log.rename(columns = {activity_key: 'concept:name'}, inplace = True)
        if timestamp_key != 'time:timestamp':
            log.rename(columns = {timestamp_key: 'time:timestamp'}, inplace = True)
        self.categorical_casefeatures = categorical_casefeatures
        self.numerical_casefeatures = numerical_casefeatures
        self.categorical_eventfeatures = categorical_eventfeatures
        self.numerical_eventfeatures = numerical_eventfeatures
        self.outcome = outcome
        
        # all_event_features = categorical_eventfeatures + numerical_eventfeatures
        print("Preprocessing the data...")
        self.log = _preprocess_pipeline(log = log, start_date= start_date, end_date = end_date) 
        self.original_log = self.log.copy()
        # Getting the list of all Directly-Follows Relations (DFRs) present in the log
        self.dfrelations, self.dfr_df = get_sorted_DFRs(self.log)

        # Getting the (sorted) list of all variants present in the log
        case_variant = get_variant_case(self.log)
        self.all_vars = list(case_variant['variant'].value_counts().index)
        self.variant_df = pd.DataFrame(case_variant['variant'].value_counts()).reset_index()
        self.variant_df.columns = ['variant', 'variant count']



    # Plotting methods
        
    def topK_dfr_evol(self, time_unit = 'days', frequency = 'weekly', case_assignment = 'first_event', plt_type = 'univariate', numeric_agg = 'mean', max_k = 10, xtr_outlier_rem = True):
        """Plot the time series of the requested aggregations for each of the ``max_k`` 
        most frequently occurring Directly-Follows Relations (DFRs).

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. The requested DFR evolutions plotted are arranged in 
        descending order of the number of occurrences of the DFRs.

        For readability, the DFRs are encoded with a number with DFR 1 being the most 
        frequently occurring DFR, and DFR ``max_k`` being the max_k'th most occurring 
        DFR. To retrieve a dataframe that maps these DFR numbers to the actual 
        activity pairs, the ``get_DFR_df()`` method of the ``DynamicLogPlots`` class can 
        be called upon. 

        Parameters
        ----------
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome', 'type_dfr_performance'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified 
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        max_k : int, optional
            Only the max_k most frequently occurring DFRs are considered, by default 10. 
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular 
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.

        Notes
        -----
        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs. 

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval. 

        * ``plt_type`` can take one of the following five values:

          * `'univariate'`: For each DFR, the evolutions of the periodically aggregated 
            amount of occurrences per case, as well as the fraction of cases containing
            at least one occurance of of that DFR, are plotted. 

          * `'type_tt'`: For each DFR, next to the univariate plots, also the evolutions of 
            the periodically aggregated Throughput Time (TT) for cases with vs. cases 
            without that DFR are plotted. 

          * `'type_events_case'`: For each DFR, next to the univariate plots, also the evolutions of 
            the periodically aggregated case length (in number of events per case (NEPC)) for cases 
            with vs. cases without that DFR are plotted. 

          * `'type_outcome'`: For each DFR, next to the univariate plots, also the evolutions of 
            the periodic fractions of cases with a positive outcome ('outcome=1') for cases with
            vs. cases without that DFR, are plotted. (Only applicable if an outcome is already 
            specified upon initialization of the `DynamicLogPlots` instance, or with the 
            ``set_outcomeColumn(outcome)`` method.)

          * `'type_dfr_performance'`: For each DFR, next to the univariate plots, also the 
            evolution of the periodically aggregated DFR performance is plotted. The DFR 
            performance refers to the time elapsed between the first and last activity of 
            that DFR. The time unit in which these periodic performance aggregations are 
            expressed, is automatically determined based on their magnitude. **NOTE** that 
            the `'type_dfr_performance'` representation type is a special case, as it is 
            the only representation type in which cases can deliver more than one measure 
            for a certain DFR. I.e. cases that contain more than one occurrence of a 
            certain DFR will also deliver more than one value for the DFR performance to 
            be aggregated over. 

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='univariate'`, for each time interval, 
            and for each of the `max_k` DFRs, the mean amount of occurrences per case 
            is computed. The fraction of cases containing at least one occurrence of a 
            certain DFR is simply computed by, for each time interval, computing the 
            amount of cases with at least one occurrence, and dividing it by the total 
            amount of cases assigned to that time interval. 

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` DFRs, the median amount of occurrences per case, 
            the median TT for cases with at least one occurrence of that DFR, and the 
            median TT for all other cases (without an occurrence of that DFR) are 
            computed. 

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            and for each of the `max_k` DFRs, the minimum amount of occurrences per case, 
            the minimum NEPC for cases with at least one occurrence of that DFR, and the 
            minimum NEPC for all other cases (without an occurrence of that DFR) are 
            computed. 

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_dfr_performance'`, for each 
            time interval, and for each of the `max_k` DFRs, the maximum amount of 
            occurrences per case and the maximum DFR performance (i.e. the maximum 
            time elapsed between the occurrence of the first and second activity of 
            that DFR) are computed. 

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_dfr_performance'`, 
            for each time interval, and for each of the `max_k` DFRs, the standard deviation 
            of amount of occurrences per case as and the standard deviation of the DFR 
            performances are computed. 

        """
        valm._verify_topKdfr(outcome = self.outcome, time_unit = time_unit, frequency= frequency, 
                            case_assignment= case_assignment, plt_type= plt_type, numeric_agg= numeric_agg,
                             max_k= max_k, xtr_outlier_rem = xtr_outlier_rem)
        top_k_dfr = self.dfrelations[:max_k]
        up.topK_dfr_evol(log = self.log, top_k_dfr = top_k_dfr, outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment,
                        type = plt_type, numeric_agg = numeric_agg, max_k = max_k, xtr_outlier_rem = xtr_outlier_rem)
    
    def dfr_evol(self, directly_follows_relations, time_unit='days', frequency='weekly', case_assignment = 'first_event', plt_type= 'univariate', numeric_agg= 'mean', xtr_outlier_rem=True):
        """Plot the time series of the requested aggregations for the 
        Directly-Follows Relations (DFRs) specified in the ``directly_follows_relations`` list.

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. The requested DFR evolutions plotted are arranged in 
        descending order of the number of occurrences of the DFRs.

        For readability, the DFRs in ``directly_follows_relations`` are encoded with a 
        number, with DFR 1 being the most frequently occurring DFR, DFR 2 being 
        the second most occurring DFR, and so on. To retrieve a dataframe that maps these 
        DFR numbers to the actual activity pairs, the ``get_DFR_df()`` method of the 
        ``DynamicLogPlots`` class can be called upon. 

        Parameters
        ----------
        directly_follows_relations : list of tuple
            The DFRs for which the requested time series will be plotted. Each DFR needs to 
            be specified as a tuple that contains 2 strings, referring to the 2 activities in 
            the DFR, e.g. ('activity_a', 'activity_b').
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome', 'type_dfr_performance'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular  
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.

        Notes
        -----
        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following five values:

          * `'univariate'`: For each DFR, the evolutions of the periodically aggregated 
            amount of occurrences per case, as well as the fraction of cases containing
            at least one occurance of of that DFR, are plotted.

          * `'type_tt'`: For each DFR, next to the univariate plots, also the evolutions of 
            the periodically aggregated Throughput Time (TT) for cases with vs. cases 
            without that DFR are plotted.

          * `'type_events_case'`: For each DFR, next to the univariate plots, also the evolutions of 
            the periodically aggregated case length (in number of events per case (NEPC)) for cases 
            with vs. cases without that DFR are plotted.

          * `'type_outcome'`: For each DFR, next to the univariate plots, also the evolutions of 
            the periodic fractions of cases with a positive outcome ('outcome=1') for cases with
            vs. cases without that DFR, are plotted. (Only applicable if an outcome is already 
            specified upon initialization of the `DynamicLogPlots` instance, or with the 
            ``set_outcomeColumn(outcome)`` method.)

          * `'type_dfr_performance'`: For each DFR, next to the univariate plots, also the 
            evolution of the periodically aggregated DFR performance is plotted. The DFR 
            performance refers to the time elapsed between the first and last activity of
            that DFR. The time unit in which these periodic performance aggregations are 
            expressed, is automatically determined based on their magnitude. **NOTE** that 
            the `'type_dfr_performance'` representation type is a special case, as it is 
            the only representation type in which cases can deliver more than one measure 
            for a certain DFR. I.e. cases that contain more than one occurrence of a 
            certain DFR will also deliver more than one value for the DFR performance to 
            be aggregated over.
        
        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='univariate'`, for each time interval, 
            and for each of the DFRs specified in `directly_follows_relations`, the 
            mean amount of occurrences per case is computed. The fraction of cases 
            containing at least one occurrence of a certain DFR is simply computed by, 
            for each time interval, computing the amount of cases with at least one 
            occurrence, and dividing it by the total amount of cases assigned to 
            that time interval.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the DFRs specified in `directly_follows_relations`, the median 
            amount of occurrences per case, the median TT for cases with at least one 
            occurrence of that DFR, and the median TT for all other cases 
            (without an occurrence of that DFR) are computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            and for each of the DFRs specified in `directly_follows_relations`, the 
            minimum amount of occurrences per case, the minimum NEPC for cases with 
            at least one occurrence of that DFR, and the minimum NEPC for all other 
            cases (without an occurrence of that DFR) are computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_dfr_performance'`, for each 
            time interval, and for each of the DFRs specified in `directly_follows_relations`, 
            the maximum amount of occurrences per case and the maximum DFR performance 
            (i.e. the maximum time elapsed between the occurrence of the first and second 
            activity of that DFR) are computed.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_dfr_performance'`, 
            for each time interval, and for each of the DFRs specified in `directly_follows_relations`, 
            the standard deviation of amount of occurrences per case as and the standard 
            deviation of the DFR performances are computed.

        """

        valm._verify_dfr(dfrelations = self.dfrelations, directly_follows_relations = directly_follows_relations, outcome = self.outcome, 
                        time_unit = time_unit, frequency = frequency, case_assignment = case_assignment, plt_type = plt_type, 
                        numeric_agg = numeric_agg, xtr_outlier_rem= xtr_outlier_rem)

        up.dfr_evol(   log = self.log, directly_follows_relations = directly_follows_relations, outcome = self.outcome, time_unit = time_unit, frequency = frequency,
                    case_assignment = case_assignment, type = plt_type, numeric_agg = numeric_agg, xtr_outlier_rem = xtr_outlier_rem)


    def topK_variants_evol(self, time_unit='days', frequency='weekly', case_assignment = 'first_event', plt_type= 'univariate', numeric_agg= 'mean', max_k= 10, xtr_outlier_rem = True):
        """Plot the time series of the requested aggregations for each of the ``max_k`` 
        most frequently occurring variants.

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. The requested variant time series plotted are arranged in 
        descending order of the number of occurrences of the variants.

        For readability, the variants are encoded with a number, with variant 1 being the most 
        frequently occurring variant, and variant ``max_k`` being the max_k'th most occurring 
        variant. To retrieve a dataframe that maps these variant numbers to the actual 
        activity sequences, the ``get_var_df()`` method of the ``DynamicLogPlots`` class can 
        be called upon. 

        Parameters
        ----------
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        max_k : int, optional
            Only the max_k most frequently occurring variants are considered, by default 10. 
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular 
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.

        Notes
        -----

        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs. 
          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following three values:

          * `'univariate'`: For each variant, plots the evolution of the periodic fraction
            of cases for accounted for by that variant.

          * `'type_tt'`: For each variant, next to the univariate plots, also the evolutions of 
            the periodically aggregated Throughput Time (TT) for cases belonging to that variant vs. 
            all other cases (not belonging to that variant), are plotted.

          * `'type_outcome'`: For each variant, next to the univariate plots, also the evolutions 
            of the periodic fractions of cases with a positive outcome ('outcome=1') for cases 
            belonging to that variant vs. all other cases (not belonging to that variant),
            are plotted. (Only applicable if an outcome is already specified upon initialization 
            of the `DynamicLogPlots` instance, or with the ``set_outcomeColumn(outcome)`` method.)

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` variants, the mean TT for cases belonging to 
            that variant, and the mean TT for all other cases (belonging to other 
            variants) are computed.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` variants, the median TT for cases belonging to 
            that variant, and the median TT for all other cases (belonging to other 
            variants) are computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` variants, the minimum TT among cases belonging 
            to that variant, and the minimum TT among all other cases (belonging to other 
            variants) are computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` variants, the maximum TT among cases belonging 
            to that variant, and the maximum TT among all other cases (belonging to other 
            variants) are computed.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_tt'`, 
            for each time interval, and for each of the `max_k` variants, the standard  
            deviation of the TT for cases belonging to that variant, and the standard 
            deviation of the TT for all other cases (belonging to other variants) are computed.

        """
        valm._verify_topKvariants(outcome = self.outcome, time_unit = time_unit, frequency = frequency, 
                                    case_assignment = case_assignment, plt_type = plt_type, numeric_agg = numeric_agg,
                                    max_k = max_k, xtr_outlier_rem = xtr_outlier_rem)
        top_k_vars = self.all_vars[:max_k]
        up.topK_variants_evol(log = self.log, top_k_vars = top_k_vars, outcome = self.outcome, time_unit= time_unit, frequency= frequency, 
                            case_assignment = case_assignment, type= plt_type, numeric_agg= numeric_agg, max_k= max_k, xtr_outlier_rem = xtr_outlier_rem)

    def variants_evol(self, variants, time_unit='days', frequency='weekly', case_assignment = 'first_event', plt_type= 'univariate', numeric_agg= 'mean', xtr_outlier_rem = True):
        """Plot the time series of the requested aggregations for the 
        variants specified in the ``variants`` list.

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. The requested variant time series plotted are arranged in 
        descending order of the number of occurrences of the variants.

        For readability, the variants are encoded with a number, with variant 1 being the most 
        frequently occurring variant, variant 2 being the second most occurring 
        variant, and so on. To retrieve a dataframe that maps these variant numbers to the actual 
        activity sequences, the ``get_var_df()`` method of the ``DynamicLogPlots`` class can 
        be called upon. 

        Parameters
        ----------
        variants : list of tuple
            The variants for which the requested time series will be plotted. Each variant needs to 
            be specified as a tuple that contains N strings, referring to the N activities that 
            constitute that variant. 
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular  
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.

        Notes
        -----
        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following three values:

          * `'univariate'`: For each variant, plots the evolution of the periodic fraction
            of cases accounted for by that variant.

          * `'type_tt'`: For each variant, next to the univariate plots, also the evolutions of 
            the periodically aggregated Throughput Time (TT) for cases belonging to that variant vs. 
            all other cases (not belonging to that variant), are plotted.

          * `'type_outcome'`: For each variant, next to the univariate plots, also the evolutions 
            of the periodic fractions of cases with a positive outcome ('outcome=1') for cases 
            belonging to that variant vs. all other cases (not belonging to that variant),
            are plotted. (Only applicable if an outcome is already specified upon initialization 
            of the `DynamicLogPlots` instance, or with the ``set_outcomeColumn(outcome)`` method.)

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each variant specified in `variants`, the mean TT for cases belonging to 
            that variant, and the mean TT for all other cases (belonging to other 
            variants) are computed.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each variant specified in `variants`, the median TT for cases belonging to 
            that variant, and the median TT for all other cases (belonging to other 
            variants) are computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each variant specified in `variants`, the minimum TT among cases belonging 
            to that variant, and the minimum TT among all other cases (belonging to other 
            variants) are computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each variant specified in `variants`, the maximum TT among cases belonging 
            to that variant, and the maximum TT among all other cases (belonging to other 
            variants) are computed.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_tt'`, 
            for each time interval, and for each variant specified in `variants`, the standard  
            deviation of the TT for cases belonging to that variant, and the standard 
            deviation of the TT for all other cases (belonging to other variants) are computed.

        """
        valm._verify_variants(all_vars = self.all_vars, variants = variants, outcome = self.outcome, time_unit = time_unit, 
                                frequency = frequency, case_assignment = case_assignment, plt_type = plt_type, 
                                numeric_agg = numeric_agg, xtr_outlier_rem = xtr_outlier_rem)

        up.variants_evol(log = self.log, variants = variants, outcome = self.outcome, time_unit = time_unit,
                            frequency = frequency, case_assignment = case_assignment, type= plt_type, 
                            numeric_agg= numeric_agg, xtr_outlier_rem = xtr_outlier_rem)

    def topK_categorical_caseftr_evol(self, case_feature, time_unit = 'days', frequency = 'weekly', case_assignment = 'first_event', plt_type = 'univariate', numeric_agg = 'mean', max_k = 10, xtr_outlier_rem = True):
        """Plot the time series of the requested aggregations for each of the ``max_k`` 
        most frequently occurring levels of categorical case feature 'case_feature'.
        If ``max_k`` is greater than or equal to the cardinality of that feature, the 
        requested time series for all levels are plotted. 

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. The requested level time series plotted are arranged in 
        descending order of the number of occurrences of the levels. 

        Parameters
        ----------
        case_feature : str
            Column name of the categorical case feature. The case feature already has to be 
            specified, either upon initalization of the `DynamicLogPlots` object, or with 
            the `.add_categorical_caseft(case_feature)` class method. 
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        max_k : int, optional
            Only the max_k most frequently occurring levels of 'case_feature' are considered, 
            by default 10.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular 
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.

        Notes
        -----

        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following four values:

          * `'univariate'`: For each level, plots the evolution of the periodic fraction
            of cases for which `case_feature='level'`.

          * `'type_tt'`: For each level, next to the univariate plots, also the evolutions of 
            the periodically aggregated Throughput Time (TT) for cases with `case_feature='level'` vs. 
            all other cases (`case_feature!='level'`), are plotted.

          * `'type_events_case'`: For each level, next to the univariate plots, also the evolutions of 
            the periodically aggregated case length (in Number of Events Per Case (NEPC)) for cases 
            with `case_feature='level'` vs. all other cases (`case_feature!='level'`), are plotted.

          * `'type_outcome'`: For each level, next to the univariate plots, also the evolutions 
            of the periodic fractions of cases with a positive outcome ('outcome=1') for cases 
            with `case_feature='level'` vs. all other cases (`case_feature!='level'`), are plotted. 
            (Only applicable if an outcome is already specified upon initialization 
            of the `DynamicLogPlots` instance, or with the ``set_outcomeColumn(outcome)`` method.)

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` levels of `case_feature`, the mean TT for cases with 
            `case_feature='level'`, and the mean TT for all other cases (with 
            `case_feature!='level'`) are computed.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            and for each of the `max_k` levels of `case_feature`, the median NEPC for cases with 
            `case_feature='level'`, and the median NEPC for all other cases (with 
            `case_feature!='level'`) are computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            and for each of the `max_k` levels of `case_feature`, the minimum NEPC among cases with 
            `case_feature='level'`, and the minimum NEPC among all other cases (with 
            `case_feature!='level'`) are computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` levels of `case_feature`, the maximum TT among cases 
            with `case_feature='level'`, and the maximum TT among all other cases (with 
            `case_feature!='level'`) are computed.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` levels of `case_feature`, the standard deviation of the 
            TT among cases with `case_feature='level'`, and the standard deviation of the TT of all 
            other cases (with `case_feature!='level'`) are computed.

        """

        valm._verify_topKcatFt(feature = case_feature, ftr_type = 'case', feature_list = self.categorical_casefeatures, 
                                    outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment, 
                                    plt_type = plt_type, numeric_agg = numeric_agg, max_k = max_k, xtr_outlier_rem = xtr_outlier_rem)

        up.topK_categorical_caseftr_evol(log = self.log, case_feature = case_feature, outcome = self.outcome, time_unit = time_unit, 
                                        frequency = frequency, case_assignment = case_assignment, type = plt_type, 
                                        numeric_agg = numeric_agg, max_k = max_k, xtr_outlier_rem = xtr_outlier_rem)


    def num_casefts_evol(self, numeric_case_list, time_unit='days', frequency='weekly', case_assignment = 'first_event', plt_type = 'univariate', numeric_agg= 'mean', xtr_outlier_rem = True):
        """ Plots the time series of the requested aggregations over time for the numerical 
        event features specified in the ``numeric_case_list`` argument.

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument.

        Parameters
        ----------
        numeric_case_list : list of str
            Column names of the numerical case features. All case features already have to be 
            specified, either upon initalization of the `DynamicLogPlots` object, or with 
            the `.add_numerical_caseft(case_feature)` class method. 
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular  
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.

        Notes
        -----

        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following four values:

          * `'univariate'`: For each numerical case feature, plots the evolution of its 
            periodical aggregations.

          * `'type_tt'`: For each numerical case feature, next to the univariate plots, also the
            evolution of the periodically aggregated ratio of Throughput Time (TT) needed per 
            unit of that feature is plotted. The time unit of these periodic ratio aggregations 
            (time unit / unit of feature) is automatically determined based on their magnitude.

          * `'type_events_case'`: For each numerical case feature, next to the univariate plots, 
            also the evolution of the periodically aggregated ratio of Number of Events Per Case 
            (NEPC) needed per 10^(x) units of that feature is plotted. The exponent 'x' is automatically 
            determined based on their magnitude.

          * `'type_outcome'`: For each numerical case feature, next to the univariate plots, also 
            the two evolutions of the feature's periodical aggregations for cases with a positive 
            outcome ('outcome=1') vs. cases with a negative outcome ('outcome=1') are plotted. 
            (Only applicable if an outcome is already specified upon initialization 
            of the `DynamicLogPlots` instance, or with the ``set_outcomeColumn(outcome)`` method.)

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='univariate'`, for each time interval, 
            and for each numeric case feature specified in `numeric_case_list`, the mean 
            of all case feature values is is computed.

          * `'median'` : The requested time series are computed by taking the median for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each numeric case feature specified in `numeric_case_list`, the median 
            of all case feature values as well as the median of all TT ratios are computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_events_case'`, for each time 
            interval, and for each numeric case feature specified in `numeric_case_list`, 
            both the minimum of all case feature values and the minimum of all NEPC ratios 
            are computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_outcome'`, for each time 
            interval, and for each numeric case feature specified in `numeric_case_list`, 
            the maximum of all case feature values, the maximum of all case feature values 
            for cases with a positive outcome and the maximum of all values for cases 
            with a negative outcome are computed.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_tt'`, for each 
            time interval, and for each numeric case feature specified in `numeric_case_list`, 
            the standard deviation of all case feature values as well as the standard deviation 
            of all TT ratios are computed.

        """

        valm._verify_numFt(features = numeric_case_list, ftr_type = 'case', feature_list = self.numerical_casefeatures, 
                            outcome = self.outcome, time_unit = time_unit, frequency = frequency, 
                            case_assignment= case_assignment, plt_type = plt_type, numeric_agg = numeric_agg, xtr_outlier_rem= xtr_outlier_rem)
        up.num_casefts_evol(log = self.log, numeric_case_list = numeric_case_list, outcome = self.outcome, time_unit = time_unit, 
                                frequency = frequency, case_assignment = case_assignment, type = plt_type, 
                                numeric_agg= numeric_agg, xtr_outlier_rem= xtr_outlier_rem)


    def topK_categorical_eventftr_evol(self, event_feature, time_unit = 'days', frequency = 'weekly', case_assignment = 'first_event', plt_type = 'univariate', numeric_agg = 'mean', max_k = 10, xtr_outlier_rem = True):
        """Plot the time series of the requested aggregations for each of the ``max_k`` 
        most frequently occurring levels of categorical event feature ``event_feature``.
        If ``max_k`` is greater than or equal to the cardinality of that feature, the 
        requested time series for all levels are plotted. 

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. The requested level time series plotted are arranged in 
        descending order of the number of occurrences of each level.

        Parameters
        ----------
        event_feature : str
            Column name of the categorical event feature. The event feature already has to be 
            specified, either upon initalization of the `DynamicLogPlots` object, or with 
            the `.add_categorical_eventft(event_feature)` class method. 
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        max_k : int, optional
            Only the max_k most frequently occurring levels of 'event_feature' are considered, 
            by default 10.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular  
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.

        Notes
        -----

        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following four values:

          * `'univariate'`: For each level, plots the evolution of the periodic fraction
            of cases for which at least one event satisfies `event_feature='level'`. 
          * `'type_tt'`: For each level, next to the univariate plots, also the evolutions of 
            the periodically aggregated Throughput Time (TT) for cases with at least one  
            occurrence of `event_feature='level'` vs. all other cases (`event_feature!='level'` for all 
            events), are plotted.

          * `'type_events_case'`: For each level, next to the univariate plots, also the evolutions of 
            the periodically aggregated case length (in Number of Events Per Case (NEPC)) for cases 
            with at least one occurrence of `event_feature='level'` vs. all other cases 
            (`event_feature!='level'` for all events), are plotted.

          * `'type_outcome'`: For each level, next to the univariate plots, also the evolutions 
            of the periodic fractions of cases with a positive outcome ('outcome=1') for cases 
            with at least one occurrence of `event_feature='level'` vs. all other cases 
            (`event_feature!='level'` for all events), are plotted. (Only applicable if an outcome 
            is already specified upon initialization of the `DynamicLogPlots` instance, or with 
            the ``set_outcomeColumn(outcome)`` method.)

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` levels of `event_feature`, the mean TT for cases with 
            at least one occurrence of `event_feature='level'`, and the mean TT for all other cases 
            (with `event_feature!='level'` for all events) are computed.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            and for each of the `max_k` levels of `event_feature`, the median NEPC for cases with 
            at least one occurrence of `event_feature='level'`, and the median NEPC for all other cases  
            (with `event_feature!='level'` for all events) are computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            and for each of the `max_k` levels of `event_feature`, the minimum NEPC among cases with 
            at least one occurrence of `event_feature='level'`, and the minimum NEPC among all other 
            cases (with `event_feature!='level'` for all events) are computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` levels of `event_feature`, the maximum TT among cases 
            with at least one occurrence of `event_feature='level'`, and the maximum TT among all 
            other cases (with `event_feature!='level'` for all events) are computed.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each of the `max_k` levels of `event_feature`, the standard deviation of the 
            TT among cases with at least one occurrence of `event_feature='level'`, and the standard 
            deviation of the TT of all other cases (with `event_feature!='level'` for all events) are 
            computed.

        """
        valm._verify_topKcatFt(feature = event_feature, ftr_type = 'event', feature_list = self.categorical_eventfeatures,
                                outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment,
                                plt_type = plt_type, numeric_agg = numeric_agg, max_k = max_k, xtr_outlier_rem = xtr_outlier_rem)

        up.topK_categorical_eventftr_evol(log = self.log, event_feature = event_feature, outcome = self.outcome, time_unit = time_unit,
                                frequency = frequency, case_assignment = case_assignment, plt_type = plt_type, numeric_agg = numeric_agg,
                                max_k = max_k, xtr_outlier_rem = xtr_outlier_rem)


    def num_eventfts_evol(self, numeric_event_list, time_unit = 'days', frequency = 'weekly', case_assignment = 'first_event', 
                          plt_type = 'univariate', numeric_agg = 'mean', xtr_outlier_rem = True, numEventFt_transform = 'last'):
        """Plot the evolution of the requested aggregations over time for the numerical 
        event features specified in the ``numeric_event_list`` argument. 

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. 

        In contrast to the numerical case features, numerical event features can take on 
        different values for one and the same case. Therefore, to visualize case-level 
        characteristics over time, an additional abstraction method is needed to project a 
        trace's sequence of numerical event feature values to a single numeric value. This
        transformation can be specified by means of the ``numEventFt_transform`` parameter. 
        
        Parameters
        ----------
        numeric_event_list : list of str
            Column names of the numerical event features. All event features already have to be 
            specified, either upon initalization of the `DynamicLogPlots` object, or with 
            the `.add_numerical_eventft(event_feature)` class method. 
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular  
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.
        numEventFt_transform : {'last', 'first', 'mean', 'median', 'sum', 'prod', 'min', 'max'}
            If any numeric event features contained in 'event_features', `numEventFt_transform` 
            determines the way in which these numerical event features are transformed to the 
            case level. By default `'last'`. For a more detailed explanation of the different 
            `numEventFt_transform` options, see Notes.

        
        Notes
        -----

        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following four values:

          * `'univariate'`: For each numerical event feature, plots the evolution of its 
            periodical aggregations.

          * `'type_tt'`: For each numerical event feature, next to the univariate plots, also the
            evolution of the periodically aggregated ratio of Throughput Time (TT) needed per 
            unit of that feature is plotted. The time unit of these periodic ratio aggregations 
            (time unit / unit of feature) is automatically determined based on their magnitude.

          * `'type_events_case'`: For each numerical event feature, next to the univariate plots, 
            also the evolution of the periodically aggregated ratio of Number of Events Per Case 
            (NEPC) needed per 10^(x) units of that feature is plotted. The exponent 'x' is automatically 
            determined based on their magnitude.

          * `'type_outcome'`: For each numerical event feature, next to the univariate plots, also 
            the two evolutions of the feature's periodical aggregations for cases with a positive 
            outcome ('outcome=1') vs. cases with a negative outcome ('outcome=1') are plotted. 
            (Only applicable if an outcome is already specified upon initialization 
            of the `DynamicLogPlots` instance, or with the ``set_outcomeColumn(outcome)`` method.)

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval. E.g. for `plt_type='univariate'`, for each time interval, 
            and for each numeric event feature, the mean of all (transformed) values is 
            is computed.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            and for each numeric event feature, the median of all (transformed) event 
            feature values as well as the median of all TT ratios are computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_events_case'`, for each time  
            interval, and for each numeric event feature, both the minimum of all 
            (transformed) values and the minimum of all NEPC ratios are computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_outcome'`, for each time  
            interval, and for each numeric event feature, both the maximum of all 
            (transformed) values, the maximum of all (transformed) values for cases 
            with a positive outcome and the maximum of all (transformed) values for 
            cases with a negative outcome are computed.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval. E.g. for `plt_type='type_tt'`, for each 
            time interval, and for each numeric event feature, the standard deviation 
            of all (transformed) event feature values as well as the standard deviation 
            of all TT ratios are computed.

        * 'numEventFt_transform' determines, for each numerical event feature contained in 
          `numeric_event_list`, how each trace's sequence of numerical event feature values 
          are projected to a single numeric value. It can take on the following eight values:

          * `'last'` : Each case is assigned the last non-null entry a numerical event feature.

          * `'first'` : Each case is assigned the first non-null entry of a numerical event feature.

          * `'mean'` : Each case is assigned the mean value over all its non-null entries of a 
            numerical event feature.

          * `'median'` : Each case is assigned the median value over all its non-null entries of a 
            numerical event feature.

          * `'sum'` : Each case is assigned the sum over all its non-null entries of a 
            numerical event feature.

          * `'prod'` : Each case is assigned the product over all its non-null entries of a 
            numerical event feature.

          * `'min'` : Each case is assigned the minimum value over all its non-null entries of a 
            numerical event feature.

          * `'max'` : Each case is assigned the maximum value over all its non-null entries of a 
            numerical event feature.

        """

        valm._verify_numFt(features = numeric_event_list, ftr_type = 'event', feature_list = self.numerical_eventfeatures,
                            outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment,
                            plt_type = plt_type, numeric_agg = numeric_agg, xtr_outlier_rem = xtr_outlier_rem, 
                            numEventFt_transform = numEventFt_transform)
                            
        up.num_eventfts_evol(log = self.log, numeric_event_list = numeric_event_list, outcome = self.outcome, time_unit = time_unit,
                            frequency = frequency, case_assignment = case_assignment, type = plt_type, numeric_agg = numeric_agg, 
                            xtr_outlier_rem = xtr_outlier_rem, numEventFt_transform = numEventFt_transform)


    def distinct_variants_evol(self, time_unit='days', frequency='weekly', case_assignment = 'first_event', plt_type= 'univariate', numeric_agg= 'mean', xtr_outlier_rem = True, cases_initialized = True):
        """
        **Deprecated**

        .. deprecated:: 0.1.1
          ``distinct_variants_evol()`` method will be removed in DyLoPro 0.2.0 
          because it duplicates part of the visualization capabilities 
          offered by the ``distinct_variants_AdvancedEvol()`` method. 
          Use the more extensive ``distinct_variants_AdvancedEvol()`` 
          visualization method instead.
        
        Plots the number of distinct variants over time, as well as the number of 
        distinct previously unseen / distinct new variants over time.

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        combined with different performance measures by specifying the 
        ``plt_type`` argument.

        Parameters
        ----------
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular  
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.
        cases_initialized : bool, optional
            If True, and `plt_type!='univariate'`, then the method also plots 
            the evolution of the number of cases initialized in each period 
            on the same graph as the one foreseen for either the Throughput Time 
            (`plt_type='type_tt'`), the Number of Events Per Case (`plt_type='type_events_case'`) 
            or the fraction of positive cases (`plt_type='type_outcome'`), by default `True`.

        Notes
        -----
        
        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.
            
        * ``plt_type`` can take one of the following four values:

          * `'univariate'`: Plots the evolution of the amount of distinct variants present in each time period, 
            as well as the amount of distinct new variants introduced in each time period.

          * `'type_tt'`: Next ot the univariate plots, also plots the evolution of the periodically aggregated 
            Throughput Time (TT) (in the time unit specified by the 'time_unit' argument).

          * `'type_events_case'`: Next to the univariate plots, also plots the evolution of the periodically aggregated 
            Number of Events Per Case (NEPC).

          * `'type_outcome'`: Next to the univariate plots, also plots the evolution of the periodically aggregated 
            fraction of cases with outcome = True. (Only applicable if an outcome is already specified upon initialization 
            of the `DynamicLogPlots` instance, or with the ``set_outcomeColumn(outcome)`` method.)

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time interval.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval.

          * `'std'` : The requested time series are computed by taking the standard 
            deviation for each time interval.

        """
        valm._verify_distinctvars(outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment, plt_type = plt_type, 
                                    numeric_agg = numeric_agg, xtr_outlier_rem = xtr_outlier_rem, cases_initialized = cases_initialized)
        
        up.distinct_variants_evol(log = self.log, outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment, 
                                    type = plt_type, numeric_agg = numeric_agg, xtr_outlier_rem = xtr_outlier_rem, cases_initialized = cases_initialized)


    def distinct_variants_AdvancedEvol(self, time_unit='days', frequency='weekly', case_assignment = 'first_event', plt_type= 'univariate', numeric_agg= 'mean', xtr_outlier_rem = True, cases_initialized = True):
        """Plot the evolution of the requested aggregations over time for:

        #. cases belonging to variants already seen in previous time periods
        #. cases belonging to variants first observed in each time period

        All cases are grouped into time intervals of which the length is determined by the 
        ``frequency`` argument. The condition that determines the time interval to which a 
        certain case is assigned, is determined by the ``case_assignment`` argument. Can be 
        holistically combined with different performance measures by specifying the 
        ``plt_type`` argument. 

        Parameters
        ----------
        time_unit : {'microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks'}
            Time unit in which the throughput time of cases is specified, by default `'days'`.
        frequency : {'minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly' '2-hourly', 
                    '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 'quarterly', 'half-yearly'}
            Frequency by which the observations are grouped together, by default `'weekly'`.
        case_assignment : {'first_event', 'last_event', 'max_events'}
            Determines the condition upon which each case is assigned to a certain 
            period, by default `'first_event'`. For a more detailed explanation 
            of the different `case_assignment` options, see Notes.
        plt_type : {'univariate', 'type_tt', 'type_events_case', 'type_outcome'}
            Determines which time series are constructed and visualized, by default `'univariate'`.
            For a more detailed explanation of the different `plt_type` options, see Notes.
        numeric_agg : {'mean', 'median', 'min', 'max', 'std'}
            Determines how periodic quantities are aggregated, by default `'mean'`. The specified
            aggregation function will be applied to all the requested time series, except for 
            those quantities that express fractions or counts (if any). For a more detailed 
            explanation of the different `numeric_agg` options, see Notes.
        xtr_outlier_rem : bool, optional
            If True, the vertical ranges of the plots are only determined by regular  
            values, i.e. extreme outliers (>q3 + 3*iqr) in the time series are neglected 
            when determining the vertical range, by default `True`.
        cases_initialized : bool, optional
            If True, and `plt_type!='univariate'`, then the method also plots 
            the evolution of the number of cases initialized in each period 
            on the same graph as the one foreseen for either the Throughput Time 
            (`plt_type='type_tt'`), the Number of Events Per Case (`plt_type='type_events_case'`) 
            or the fraction of positive cases (`plt_type='type_outcome'`), by default `True`.

        Notes
        -----
        
        * ``case_assignment`` can take on the following three values:

          * `'first_event'`: Each case is assigned to the time interval in which 
            its first event occurs.

          * `'last_event'`: Each case is assigned to the time interval in which 
            its last event occurs.

          * `'max_events'`: Out of all the time intervals in which the events of a 
            particular case occur, the case is assigned to the interval in which most 
            of its events occur. Ties among time intervals are broken by assigning the 
            case to the first time interval.

        * ``plt_type`` can take one of the following four values:

          * `'univariate'`: Plots the periodic evolution of

            #. the absolute amount of distinct variants present in each time period

            #. the absolute amount of distinct new variants introduced in each time period

            #. the fraction of cases belonging to variants already seen in previous time periods

            #. the fraction of cases belonging to variants first observed in each time period

          * `'type_tt'`: Next ot the univariate plots, also plots the periodic evolution of

            #. the aggregated throughput time of cases pertaining to existing variants 
               in each time period
            
            #. the aggregated throughput time of cases pertaining to new 
               variants in each time period

          * `'type_events_case'`: Next to the univariate plots, also plots 
            the periodic evolution of
          
            #. the aggregated Number of Events Per Case (NEPC) of cases 
               pertaining to existing variants in each time period
            
            #. the aggregated Number of Events Per Case (NEPC) of cases 
               pertaining to new variants in each time period

          * `'type_outcome'`: (Only applicable if an outcome is already 
            specified upon initialization of the `DynamicLogPlots` instance, 
            or with the ``set_outcomeColumn(outcome)`` method.) Next to the 
            univariate plots, also plots the periodic evolution of

            #. the fraction of cases with `outcome=True` of cases 
               pertaining to existing variants in each time period 
            
            #. the fraction of cases with `outcome=True` of cases 
               pertaining to new variants in each time period

        * ``numeric_agg`` can take on the following five values:

          * `'mean'` : The requested time series are computed by taking the mean for each 
            time period. E.g. for `plt_type='type_tt'`, for each time interval, 
            the mean throughput time (TT) of cases pertaining to already 
            existing variants, and the mean TT of cases pertaining 
            to newly introduced variants is computed.

          * `'median'` : The requested time series are computed by taking the median for  
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            the median case length (in Number of Events Per Case, aka NEPC) 
            of cases pertaining to already existing variants, and the median 
            NEPC of cases pertaining to newly introduced variants is computed.

          * `'min'` : The requested time series are computed by taking the minimum for 
            each time interval. E.g. for `plt_type='type_events_case'`, for each time interval, 
            the minimum case length (in Number of Events Per Case, aka NEPC) 
            of cases pertaining to already existing variants, and the minimum 
            NEPC of cases pertaining to newly introduced variants is computed.

          * `'max'` : The requested time series are computed by taking the maximum for 
            each time interval. E.g. for `plt_type='type_tt'`, for each time interval, 
            the maximum throughput time (TT) of cases pertaining to already 
            existing variants, and the maximum TT of cases pertaining 
            to newly introduced variants is computed.

          * `'std'` : The requested time series are computed by taking the 
            standard deviation for each time interval. E.g. for 
            `plt_type='type_tt'`, for each time interval, the standard 
            deviation of the throughput time (TT) of cases pertaining to 
            already existing variants, and the standard deviation of the TT 
            of cases pertaining to newly introduced variants is computed.

        """
        valm._verify_distinctvars(outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment, plt_type = plt_type, 
                                    numeric_agg = numeric_agg, xtr_outlier_rem = xtr_outlier_rem, cases_initialized = cases_initialized)
        
        up.distinct_variants_AdvancedEvol(log = self.log, outcome = self.outcome, time_unit = time_unit, frequency = frequency, case_assignment = case_assignment, 
                                    type = plt_type, numeric_agg = numeric_agg, xtr_outlier_rem = xtr_outlier_rem, cases_initialized = cases_initialized)


    # Utility functions to add or assign new columns to certain attributes: 

    def set_outcomeColumn(self, outcome):
        """Specify an outcome column, or change the outcome column in case 
        one has already been specified.

        Parameters
        ----------
        outcome : str
            Name of the outcome column in the log. (Has to be 
            a binary outcome column with 0 and 1's, and this
            value has to be constant over all events pertaining
            to the same case.)
        """
        valm._verify_addOutCol(self.log, outcome)
        self.outcome = outcome
        

    def add_categorical_caseft(self, case_feature):
        """Specify an additional categorical case feature, on top of the one 
        specified upon initializing your `DynamicLogPlots` instance.

        Parameters
        ----------
        case_feature : str
            Name of the case_feature column in the log. Column has to be of 
            one of the following dtypes: category, object, boolean. 
        """
        if case_feature not in self.categorical_casefeatures:
            valm._verify_addCatCaFt(self.log, case_feature)
            self.categorical_casefeatures.append(case_feature)


    def add_numerical_caseft(self, case_feature):
        """Specify an additional numerical case feature, on top of the one 
        specified upon initializing your `DynamicLogPlots` instance.

        Parameters
        ----------
        case_feature : str
            Name of the case_feature column in the log. Column has to 
            be of a numerical dtype.
        """
        if case_feature not in self.numerical_casefeatures:
            valm._verify_addNumCaFt(self.log, case_feature)
            self.numerical_casefeatures.append(case_feature)

     
    def add_categorical_eventft(self, event_feature):
        """Specify an additional categorical event feature, on top of the one 
        specified upon initializing your `DynamicLogPlots` instance.

        Parameters
        ----------
        event_feature : str
            Name of the event_feature column in the log. Column has to be of 
            one of the following dtypes: category, object, boolean. 
        """
        if event_feature not in self.categorical_eventfeatures:
            valm._verify_addCatEvFt(self.log, event_feature)
            self.categorical_eventfeatures.append(event_feature)


    def add_numerical_eventft(self, event_feature):
        """Specify an additional numerical event feature, on top of the one 
        specified upon initializing your `DynamicLogPlots` instance.

        Parameters
        ----------
        event_feature : str
            Name of the event_feature column in the log. Column has to 
            be of a numerical dtype.
        """
        if event_feature not in self.numerical_eventfeatures:
            valm._verify_addNumEvFt(self.log, event_feature)
            self.numerical_eventfeatures.append(event_feature)
            
    def select_time_range(self, start_date = None, end_date = None):
        """Select only those traces starting after start_date, and ending before end_date.

        Parameters
        ----------
        start_date : str, optional
            Start date string of format "dd/mm/YYYY", by default None
        end_date : str, optional
            End date string of format "dd/mm/YYYY", by default None
        """
        valm._verify_select_time_range(start_date, end_date)
        self.log = select_timerange(log = self.original_log, start_date = start_date, end_date = end_date)
        # Recomputing the ordered list of variants
        case_variant = get_variant_case(self.log)
        self.all_vars = list(case_variant['variant'].value_counts().index)
        self.variant_df = pd.DataFrame(case_variant['variant'].value_counts()).reset_index()
        self.variant_df.columns = ['variant', 'variant count']
        # Recomputing the ordered list of Directly-Follows Relations
        self.dfrelations, self.dfr_df = get_sorted_DFRs(self.log)
    
    # Methods to get dataframes for DFR and Variant encodings in the visuals. 

    def get_DFR_df(self, max_k = None, directly_follows_relations = None, counts = False):
        """Get a `pandas.DataFrame` containing the DFR numbers together 
        with a tuple containing the two corresponding activity labels 
        of each DFR. 

        Parameters
        ----------
        max_k : int, optional
            If specified, a dataframe containing the DFR numbers and 
            their corresponding activity pairs for the `max_k` most 
            frequently occurring DFRs is returned. By default `None`.
        directly_follows_relations : list of tuple, optional
            List of tuples containing the specified DFRs. Each DFR 
            needs to be specified as a tuple that contains 2 strings, 
            referring to the 2 activities in the DFR, e.g. 
            ('activity_a', 'activity_b'). If specified and `max_k=None`, 
            a dataframe containing the DFR numbers and corresponding 
            activity pairs for each of the specified DFRs is returned. 
            If `max_k!=None`, the DFRs specified here are ignored. 
            By default `None`.
        counts : bool, optional
            If `counts=True`, the 'DFR count' column that contains the 
            number of occurrences of each DFR over the whole event log 
            is included in the returned dataframe. 

        Returns
        -------
        dfr_df_filtered : pandas.DataFrame
            Dataframe containing 'DFR number' index and the 'DFR' column 
            containing for each requested DFR the encoded DFR index and 
            the DFR tuple containing the corresponding activity pair 
            respectively. If `counts=True`, the 'DFR count' column that 
            contains the number of occurrences of each DFR over the 
            whole event log is included too. 
        """
        valm._verify_get_DFR_df(max_k, directly_follows_relations, counts, self.dfrelations)
        dfr_df_filtered = get_filtered_dfr_df(self.dfr_df, max_k = max_k, 
                                     directly_follows_relations = directly_follows_relations, 
                                     counts = counts)
        return dfr_df_filtered
    
    def get_var_df(self, max_k = None, variants = None, counts = False):
        """Get a `pandas.DataFrame` containing the variant numbers together 
        with a tuple containing the activity label strings
        of each variant. 

        Parameters
        ----------
        max_k : int, optional
            If specified, a dataframe containing the encoded variant 
            numbers and corresponding tuple of N strings (with N the 
            number of activities of a variant) is returned for 
            each of the `max_k` most frequently occurring variants. 
            By default `None`.
        variants : list of tuple
            The variants for which the requested time series will be 
            plotted. Each variant needs to be specified as a tuple that 
            contains N strings, referring to the N activities that 
            constitute that variant. If specified and `max_k=None`, 
            a dataframe containing the encoded variant numbers and 
            corresponding tuple of N strings for each of the specified 
            DFRs is returned. If `max_k!=None`, the `variants` parameter 
            is ignored. By default `None`.
        counts : bool, optional
            If `counts=True`, the 'variant count' column that contains the 
            number of occurrences of each variant over the whole event log 
            is included in the returned dataframe. 

        Returns
        -------
        filtered_var_df : pandas.DataFrame
            Dataframe containing 'variant number' index and the 'variant' 
            column containing for each requested variant the encoded variant  
            index and the variant tuple containing the corresponding activity 
            pair respectively. If `counts=True`, the 'variant count' column 
            that contains the number of occurrences of each variant over the 
            whole event log is included too. 
        """
        valm._verify_get_var_df(max_k, variants, counts, self.all_vars)
        dfr_df_filtered = get_filtered_var_df(self.variant_df, max_k = max_k, 
                                     variants = variants, 
                                     counts = counts)
        return dfr_df_filtered


import pandas as pd
import numpy as np
import datetime as dt

def _verify_ColsInLog(cols, given_cols):
    '''Verifies whether the given column names are actually present in the given log. 

        args:
            - cols: list of the names of all the columns detected in the log
            - given_cols: list of the names of the columns given by the user upon initializing the DynamicLogPlots instance. 

        returns: 
            - error_detected: boolean; True if a given column is not detected in the given log, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    '''
    error_detected, err_msge = False, " "
    for col in given_cols:
        if col not in cols:
            error_detected = True
            err_msge = "'{}': Column with column name '{}' is not found in the given log.".format(col, col)
            return error_detected, err_msge
    return error_detected, err_msge

def _verify_start_end_date(start_date, end_date):
    ''' Verifies whether the start and / or end_date are correclty specified.

        args:
            - start_date: should be a string of the following format 'dd/mm/YYYY'
            - end_date: should be a string of the following format 'dd/mm/YYYY'

        returns:
            - error_detected    : boolean; True if an error is detected in the one of the 2 dates, False otherwise.
            - err_msge          : string; returns an adequate error message if needed. 
            - err_cat           : string = 'type_error' or 'value_error'
    '''
    error_detected, err_msge, err_cat = False, " ", " "
    if start_date:
        try:
            dt.datetime.strptime(start_date, "%d/%m/%Y")
        except ValueError:
            error_detected = True
            err_msge = "'start_date' is incorrectly specified. It shoud be a string of the following format 'dd/mm/YYYY'. \
For example, November 4th 2022 should be specified as the following string: '04/11/2022'"
            err_cat = 'value_error'
            return error_detected, err_msge, err_cat
        except TypeError:
            error_detected = True
            err_msge = "'start_date' is incorrectly specified. It shoud be a string of the following format 'dd/mm/YYYY'. \
For example, November 4th 2022 should be specified as the following string: '04/11/2022'"
            err_cat = 'type_error'
            return error_detected, err_msge, err_cat

    if end_date:
        try:
            dt.datetime.strptime(end_date, "%d/%m/%Y")
        except ValueError:
            error_detected = True
            err_msge = "'end_date' is incorrectly specified. It shoud be a string of the following format 'dd/mm/YYYY'. \
For example, November 4th 2022 should be specified as the following string: '04/11/2022'"
            err_cat = 'value_error'
            return error_detected, err_msge, err_cat
        except TypeError:
            error_detected = True
            err_msge = "'end_date' is incorrectly specified. It shoud be a string of the following format 'dd/mm/YYYY'. \
For example, November 4th 2022 should be specified as the following string: '04/11/2022'"
            err_cat = 'type_error'
            return error_detected, err_msge, err_cat

    # If tests above passed, check whether start_date < end_date
    if start_date and end_date:
        dt.datetime.strptime(start_date, "%d/%m/%Y") >= dt.datetime.strptime(end_date, "%d/%m/%Y")
        error_detected = True
        err_msge = "Both 'start_date' and 'end_date' are correctly specified. However, 'start_date' ('{}') \
comes after 'end_date' ('{}').".format(start_date, end_date)
        err_cat = 'value_error'
        return error_detected, err_msge, err_cat 
    
    return error_detected, err_msge, err_cat





# outcome initialization 
def _verify_outcomeCol_initial(log, outcome, case_id_key):
    ''' Verifies whether the outcome column is correctly formatted. Is only called when a outcome column is specified,
        either when initializing the DynamicLogPlots instance, or when adding an outcome column afterwards with the 
        set_outcomeColumn(outcome) method afterwards. 

        args: 
            - log           : pd.DataFrame
            - outcome       : string referring to the outcome column of log.
            - case_id_key   : string referring to the case_id column of log.

        returns:
            - error_detected: boolean; True if the outcome column is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    '''
    error_detected = False
    err_msge = " "
    # Check whether it is a binary integer column.
    unique_vals = list(log[outcome].unique())
    if not ((unique_vals == [1, 0]) or (unique_vals == [0, 1])):
        error_detected = True
        err_msge = "Column '{}' does not have the right format. The outcome column should be a binary integer column, with the same value (0 or 1) for every event belonging to the same case.".format(outcome)

    else:
        # Check whether it is defined on case level. (Same binary value for each event of the same case)
        num_unique_percase = log.pivot_table(values = outcome, index = case_id_key, aggfunc = pd.Series.nunique)
        if len(num_unique_percase[num_unique_percase[outcome] != 1]) != 0:
            error_detected = True
            err_msge = "Column '{}' does not have the right format. The outcome column should be a binary integer column, with the same value (0 or 1) for every event belonging to the same case.".format(outcome)

    return error_detected, err_msge

# outcome check plotting methods
def _verify_outcomeCol(outcome):
    ''' Verifies whether the outcome column is already specified before including it in a plot. 

        args: 
            - outcome       : string referring to the outcome column of log.

        returns:
            - error_detected: boolean; True if the outcome column is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    '''
    # Should only be called by plotting method specific check functions when plt_type == 'type_outcome'
    error_detected = False
    err_msge = " "
    if outcome == None:
        error_detected = True
        err_msge = "'outcome = None': You want to construct plots of plt_type 'type_outcome', but you have not specified an outcome column. \
To add an outcome column to your DynamicLogPlots instance, you can call the set_outcomeColumn(outcome) method on this instance. E.g. \
<name DynamicLogPlots instance>.set_outcomeColumn('<outcome column name>')."

    return error_detected, err_msge


# plt_type checks 

#   plt_type checks for DFR-related plotting methods
def _verify_typedfr(plt_type):
    ''' Verifies whether the plt_type is correclty specified for the Directly-Follows Relations (dfr) plotting functions.

        args:
            - plt_type: string

        returns:
            - error_detected: boolean; True if plt_type is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    dfr_types = ['univariate', 'type_tt', 'type_events_case', 'type_outcome', 'type_dfr_performance']
    err_detected = False
    err_msge = " "
    if plt_type not in dfr_types:
        err_detected = True
        err_msge = "'{}' is not a valid value for the plt_type argument of this plotting method. The valid options are 'univariate' (default), \
'type_tt', 'type_events_case', 'type_outcome' or 'type_dfr_performance'.".format(plt_type)

    return err_detected, err_msge

#   plt_type checks for variant-related plotting methods 
def _verify_typevar(plt_type):
    ''' Verifies whether the plt_type is correclty specified for the Variants plotting functions.

        args:
            - plt_type: string

        returns:
            - error_detected: boolean; True if plt_type is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    err_detected = False
    err_msge = " "
    if plt_type == 'type_events_case':
        err_detected = True 
        err_msge = "While 'type_events_case' is a valid plot type for other plotting methods, it is not a valid value for the plt_type argument \
of this plotting method. The reason being: the amount of events per case for the same variant remains, by definition, constant over time.  The \
valid options are 'univariate' (default), 'type_tt' or 'type_outcome'."
        return err_detected, err_msge

    var_types = ['univariate', 'type_tt', 'type_outcome']
    if plt_type not in var_types:
        err_detected = True
        err_msge = "'{}' is not a valid value for the plt_type argument of this plotting method. The valid options are 'univariate' (default), 'type_tt' or \
'type_outcome'.".format(plt_type)

    return err_detected, err_msge

#   plt_type checks for case or event feature related plotting methods. 
def _verify_typeftrs(plt_type):
    ''' Verifies whether the plt_type is correclty specified for the (case and event) feature-related plotting functions.

        args:
            - plt_type: string

        returns:
            - error_detected: boolean; True if plt_type is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    ftr_types = ['univariate', 'type_tt', 'type_events_case', 'type_outcome']
    err_detected = False
    err_msge = " "
    if plt_type not in ftr_types:
        err_detected = True
        err_msge = "'{}' is not a valid value for the plt_type argument of this plotting method. The valid options are 'univariate' (default), 'type_tt', \
'type_events_case' or 'type_outcome'.".format(plt_type)

    return err_detected, err_msge

def _verify_typeDistVars(plt_type):
    ''' Verifies whether the plt_type is correclty specified for the distinct variant plotting method.

        args:
            - plt_type: string

        returns:
            - error_detected: boolean; True if plt_type is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    ftr_types = ['univariate', 'type_tt', 'type_events_case', 'type_outcome']
    err_detected = False
    err_msge = " "
    if plt_type not in ftr_types:
        err_detected = True
        err_msge = "'{}' is not a valid value for the plt_type argument of this plotting method. The valid options are 'univariate' (default), 'type_tt', \
'type_events_case' or 'type_outcome'.".format(plt_type)

    return err_detected, err_msge


# time_unit checks:
def _verify_time_unit(time_unit):
    ''' Verifies whether the time_unit is correclty specified for the plotting functions.

        args:
            - time_unit: string

        returns:
            - error_detected: boolean; True if time_unit is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    time_unit_list = ['microseconds', 'milliseconds', 'seconds', 'minutes', 'hours', 'days', 'weeks']
    err_detected = False
    err_msge = " "
    if time_unit not in time_unit_list:
        err_detected = True
        err_msge = "'{}' is not a valid value for the time_unit argument of the plotting methods. The valid options are 'microseconds', \
'milliseconds', 'seconds', 'minutes', 'hours', 'days' (default) or 'weeks'.".format(time_unit)

    return err_detected, err_msge

# frequency checks:
def _verify_frequency(frequency):
    ''' Verifies whether the frequency is correclty specified for the plotting functions.

        args:
            - frequency: string

        returns:
            - error_detected: boolean; True if frequency is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    frequency_list = ['minutely', '5-minutely', '10-minutely', 'half-hourly', 'hourly', 
                      '2-hourly', '12-hourly', 'daily', 'weekly', '2-weekly', 'monthly', 
                      'quarterly', 'half-yearly']
    err_detected = False
    err_msge = " "
    if frequency not in frequency_list:
        err_detected = True
        err_msge = "'{}' is not a valid value for the frequency argument of the plotting methods. The valid options are 'daily', 'weekly' \
(default), 'two_weekly' or 'monthly'.".format(frequency)

    return err_detected, err_msge

# case_assignment checks:
def _verify_case_assignment(case_assignment):
    ''' Verifies whether the case_assignment is correclty specified for the plotting functions.

        args:
            - case_assignment: string

        returns:
            - error_detected: boolean; True if case_assignment is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    case_asg_list = ['first_event', 'last_event', 'max_events']
    err_detected = False
    err_msge = " "
    if case_assignment not in case_asg_list:
        err_detected = True
        err_msge = "'{}' is not a valid value for the case_assignment argument of the plotting methods. The valid options are 'first_event' (default), \
'last_event' or 'max_events'.".format(case_assignment)

    return err_detected, err_msge

# numeric_agg checks:
def _verify_numeric_agg(numeric_agg):
    ''' Verifies whether the numeric_agg is correclty specified for the plotting functions.

        args:
            - numeric_agg: string

        returns:
            - error_detected: boolean; True if case_assignment is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    numagg_list = ['mean', 'median', 'min', 'max', 'std']
    #'mean', 'median', 'min', 'max', 'std'
    err_detected = False
    err_msge = " "
    if numeric_agg not in numagg_list:
        err_detected = True
        err_msge = "'{}' is not a valid value for the numeric_agg argument of the plotting methods. The valid options are 'mean' (default), \
'median', 'min', 'max' or 'std'.".format(numeric_agg)

    return err_detected, err_msge

# Verify numEventFt_transform argument for numerical event feature plotting method 
def _verify_numEventFt_transform(numEventFt_transform):
    numEventFt_transform_options = ['last', 'first', 'mean', 'median', 'sum', 
                                    'prod', 'min', 'max']
    err_detected, err_msge = False, " "
    if numEventFt_transform not in numEventFt_transform_options:
        err_detected = True 
        err_msge = "'{}' is not a valid value for the numEventFt_transform argument of the \
plotting method. The valid options are 'last' (default), 'first', 'mean', 'median', 'sum', \
'prod', 'min' and 'max'."

    return err_detected, err_msge 


# max_k checks:
def _verify_max_k(max_k):
    ''' Verifies whether max_k is correclty specified for the plotting functions.

        args:
            - max_k: integer (or it should be)

        returns:
            - error_detected: boolean; True if max_k is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    err_detected = False
    err_msge = " "

    if type(max_k) != int:
        err_detected = True
        err_msge = "{} is not a valid value for the 'max_k' argument of the plotting methods. 'max_k' (= 10 by default) should be an integer greater than 0.".format(max_k)
    elif max_k < 1:
        err_detected = True
        err_msge = "{} is not a valid value for the 'max_k' argument of the plotting methods. 'max_k' (=10 by default) should be an integer greater than 0.".format(max_k)

    return err_detected, err_msge

# xtr_outlier_rem checks:
def _verify_xtr_outlier_rem(xtr_outlier_rem):
    ''' Verifies whether xtr_outlier_rem is correclty specified for the plotting functions.

        args:
            - xtr_outlier_rem: bool (or it should be)

        returns:
            - error_detected: boolean; True if xtr_outlier_rem is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    err_detected = False
    err_msge = " "

    if type(xtr_outlier_rem) != bool:
        err_detected = True
        err_msge = "{} is not a valid value for the 'xtr_outlier_rem' argument of the plotting methods. \
'xtr_outlier_rem' (= True by default) should be boolean value.".format(xtr_outlier_rem)

    return err_detected, err_msge

def _verify_counts_bool(counts):
    """Verify whether counts parameter is bool dtype. 

    Parameters
    ----------
    counts : bool
    """
    err_detected = False
    err_msge = " "

    if type(counts) != bool:
        err_detected = True
        err_msge = "{} is not a valid value for the 'counts' argument. \
'counts' (= False by default) should be a boolean value.".format(counts)

    return err_detected, err_msge

# cases_initialized:
def _verify_cases_initialized(cases_initialized):
    ''' Verifies whether cases_initialized is correclty specified for the plotting functions.

        args:
            - cases_initialized: bool (or it should be)

        returns:
            - error_detected: boolean; True if xtr_outlier_rem is not well specified, False if no error is detected. 
            - err_msge      : string; returns an adequate error message if needed. 
    
    '''
    err_detected = False
    err_msge = " "

    if type(cases_initialized) != bool:
        err_detected = True
        err_msge = "{} is not a valid value for the 'cases_initialized' argument of the plotting methods. \
'cases_initialized' (= True by default) should be boolean value.".format(cases_initialized)

    return err_detected, err_msge


# Verify the passed on dfr's in the directly_follows_relations list:
def _verify_dfrelations(dfrelations: list, directly_follows_relations: list): 
    ''' Specific verification for the dfr_evol() plotting method. Verifies whether the user-defined Directly-Follows 
        Relations (DFRs), specified in the directly_follows_relations list, are correctly defined and whether they exist in the log. 

        args:
            - dfrelations                   : list of all the dfrelations present in the log.
            - directly_follows_relations    : list of user-defined dfr's to be plotted with the dfr_evol() method. 

        returns:
            - error_detected    : boolean; True if at least one user-defined DFR is specified incorrectly, or when at least one is not even
                                  present in the log.
            - err_msge          : string; returns an adequate error message if needed. 
    
    '''
    err_detected, err_msge = False, " "
    # Verify the passed on dfr's in the directly_follows_relations list:
    if type(directly_follows_relations) != list:
        err_detected = True
        err_msge = "The 'directly_follows_relations' argument is specified incorrectly. It should be specified as a list that contains the Directly-Follows \
Relations (DFRs) that need to be plotted. Note that each DFR (in the 'directly_follows_relations' list) should be represented by a tuple containing the 2 \
corresponding activity strings, e.g. ('activity_a', 'activity_b')."
        return err_detected, err_msge
    for dfrel in directly_follows_relations:
        if dfrel not in dfrelations:
            err_detected = True
            err_msge = "{}: No such Directly-Follows Relation (DFR) was found in the event log. Note that a DFR should be represented by a tuple containing \
the 2 corresponding activity strings. E.g. ('activity_a', 'activity_b').".format(dfrel)
            return err_detected, err_msge
    return err_detected, err_msge

def _verify_dfrlist_None(directly_follows_relations):
    """Verifies whether `directly_follows_relations` specified for the 
    `get_DFR_df()` method of the `DynamicLogPlots` class is not `None`, 
    in case that the method's `max_k` parameter is `None`.

    Parameters
    ----------
    directly_follows_relations : list of tuple
        List of DFRs specified by the user. 

    Returns
    -------
    err_detected : bool 
        `True` if error is detected. 
    err_msge : str
        Appropriate error message if one is detected. 
    """
    err_detected, err_msge = False, " "
    if directly_follows_relations == None:
        err_detected = True 
        err_msge = "Either the `max_k` or `directly_follows_relations` \
parameter should be specified."
    return err_detected, err_msge

def _verify_varlist_None(variants):
    """Verifies whether `variants` specified for the 
    `get_var_df()` method of the `DynamicLogPlots` class is not `None`, 
    in case that the method's `max_k` parameter is `None`.

    Parameters
    ----------
    variants : list of tuple
        List of variants specified by the user. 

    Returns
    -------
    err_detected : bool 
        `True` if error is detected. 
    err_msge : str
        Appropriate error message if one is detected. 
    """
    err_detected, err_msge = False, " "
    if variants == None:
        err_detected = True 
        err_msge = "Either the `max_k` or `variants` \
parameter should be specified."
    return err_detected, err_msge

# Verify the passed on variants in the 'variants' list:
def _verify_variantslist(all_vars: list, variants: list): 
    ''' Specific verification for the variants_evol() plotting method. Verifies whether the user-defined variants, 
        specified in the 'variants' list, are correctly defined and whether they exist in the log. 

        args:
            - all_vars  : list of all the variants present in the log.
            - variants  : list of user-defined variants to be plotted with the dfr_evol() method. 

        returns:
            - error_detected    : boolean; True if at least one user-defined variant is specified incorrectly, or when at least one is not even
                                  present in the log.
            - err_msge          : string; returns an adequate error message if needed. 
    
    '''
    err_detected, err_msge = False, " "
    # Verify the passed on variants in the 'variants' list:
    if type(variants) != list:
        err_detected = True
        err_msge = "The 'variants' argument is specified incorrectly. It should be specified as a list that contains the variants \
that need to be plotted. Note that a variant should be represented by a tuple containing the corresponding activity strings. \
For example, suppose a variant is specified by the sequence of these 3 activities: 'activity_a' - 'activity_b' - \
'activity_c', then this variant should be specified (in the variants list) as ('activity_a', 'activity_b', 'activity_c')."
        return err_detected, err_msge
    for var in variants:
        if var not in all_vars:
            err_detected = True
            err_msge = "{}: No such variant was found in the event log. Note that a variant should be represented by a tuple containing \
the corresponding activity strings. For example, suppose a variant is specified by the sequence of these 3 activities: 'activity_a' - \
'activity_b' - 'activity_c', then this variant should be specified (in the variants list) as ('activity_a', 'activity_b', 'activity_c').".format(var)
            return err_detected, err_msge
    return err_detected, err_msge


# Utility method specific checks 

#   Utility method to verify the to-be added categorical case feature 
def _verify_newCatCaFt(log, case_feature, case_id_key = 'case:concept:name'):
    ''' Validity check specifically for verifying whether the to-be added case_feature in the add_categorical_caseft(self, case_feature)
        method is valid. (Also called upon to verify the categorical case features specified upon initialization of the DynamicLogPlots class.)

        args:
            - log: pd.DataFrame 
            - case_feature: string containing the column name (in log) of the categorical case feature to-be added. 

        returns:
            - error_detected    : boolean; True if an error is detected in the to-be added categorical case feature column, False otherwise.
            - err_msge          : string; returns an adequate error message if needed. 
            - err_cat           : string = 'key_error', 'type_error' or 'value_error'
    '''
    err_detected, err_msge, err_cat = False, " ", " "
    cols = list(log.columns)
    if case_feature not in cols:
        err_detected = True
        err_msge = "'{}': Column with column name '{}' is not found in the given log.".format(case_feature, case_feature)
        err_cat = 'key_error'
        return err_detected, err_msge, err_cat

    is_obj = pd.api.types.is_object_dtype(log[case_feature])
    is_bool = pd.api.types.is_bool_dtype(log[case_feature])
    if not (is_obj or is_bool):
        err_detected = True
        err_msge = "Column '{}' is specified as a categorical case feature, but is not of the correct datatype. The categorical feature \
columns should be of the object dtype or of the boolean dtype.".format(case_feature)
        err_cat = 'type_error'
        return err_detected, err_msge, err_cat
    
    # If it passed the previous validity check, check whether it does not only contain 1 level
    num_levels = len(log[case_feature].unique())
    if num_levels == 1:
        only_value = log[case_feature].unique()[0]
        err_detected = True
        err_msge = "Column '{}' is specified as a categorical case feature, but contains only one value, namely '{}'. Hence, \
categorical case feature '{}' contains no informational value.".format(case_feature, only_value, case_feature)
        err_cat = 'value_error'
        return err_detected, err_msge, err_cat
    
    # Verify whether the case feature actually has only one constant value per case
    num_unique_percase = log.pivot_table(values = case_feature, index = case_id_key, aggfunc = pd.Series.nunique)
    if len(num_unique_percase[num_unique_percase[case_feature] > 1]) > 0:
        err_detected = True 
        err_msge = "Column '{}' is specified as a categorical case feature, but is not a valid case feature. For each individual case, \
a case feature should only have one constant value, while in feature column '{}', (some of) the cases have multiple values for this feature. \
If you are dealing with a categorical event feature, it should be added to categorical_eventfeatures.".format(case_feature, case_feature)
        err_cat = 'value_error'
    
    return err_detected, err_msge, err_cat


#   Utility method to verify the to-be added numerical case feature 
def _verify_newNumCaFt(log, case_feature, case_id_key = 'case:concept:name'):
    ''' Validity check specifically for verifying whether the to-be added case_feature in the add_numerical_caseft(self, case_feature)
        method is valid. (Also called upon to verify the numerical case features specified upon initialization of the DynamicLogPlots class.)

        args:
            - log: pd.DataFrame 
            - case_feature: string containing the column name (in log) of the numerical case feature to-be added. 

        returns:
            - error_detected    : boolean; True if an error is detected in the to-be added numerical case feature column, False otherwise.
            - err_msge          : string; returns an adequate error message if needed.
            - err_cat           : string = 'key_error', 'type_error' or 'value_error'
    '''
    err_detected, err_msge, err_cat = False, " ", " "

    # Verify whether the feature is actually in the log
    cols = list(log.columns)
    if case_feature not in cols:
        err_detected = True
        err_msge = "'{}': Column with column name '{}' is not found in the given log.".format(case_feature, case_feature)
        err_cat = 'key_error'
        return err_detected, err_msge, err_cat

    # Verify whether the numerical feature is actually of a numerical dtype.
    if not pd.api.types.is_numeric_dtype(log[case_feature]):
        err_detected = True 
        err_msge = "Column '{}' is specified as a numerical case feature, but is not of a numeric dtype.".format(case_feature)
        err_cat = 'type_error'
        return err_detected, err_msge, err_cat

    # Verify whether the case feature actually has only one constant value per case
    num_unique_percase = log.pivot_table(values = case_feature, index = case_id_key, aggfunc = pd.Series.nunique)
    if len(num_unique_percase[num_unique_percase[case_feature] > 1]) > 0:
        err_detected = True 
        err_msge = "Column '{}' is specified as a numerical case feature, but is not a valid case feature. For each individual case, \
a case feature should only have one constant value, while in feature column '{}', (some of) the cases have multiple values for this feature. \
If you are dealing with a numerical event feature, it should be added to numerical_eventfeatures.".format(case_feature, case_feature)
        err_cat = 'value_error'

    return err_detected, err_msge, err_cat
    
#   Utility method to verify the to-be added categorical event feature 
def _verify_newCatEvFt(log, event_feature):
    ''' Validity check specifically for verifying whether the to-be added event_feature in the add_categorical_eventft(self, event_feature)
        method is valid. (Also called upon to verify the categorical event features specified upon initialization of the DynamicLogPlots class.)

        args:
            - log: pd.DataFrame 
            - event_feature: string containing the column name (in log) of the categorical event feature to-be added. 

        returns:
            - error_detected    : boolean; True if an error is detected in the to-be added categorical event feature column, False otherwise.
            - err_msge          : string; returns an adequate error message if needed.
            - err_cat           : string = 'key_error' or 'type_error'
    '''
    err_detected, err_msge, err_cat = False, " ", " "
    cols = list(log.columns)
    if event_feature not in cols:
        err_detected = True
        err_msge = "'{}': Column with column name '{}' is not found in the given log.".format(event_feature, event_feature)
        err_cat = 'key_error'
        return err_detected, err_msge, err_cat
    is_obj = pd.api.types.is_object_dtype(log[event_feature])
    is_bool = pd.api.types.is_bool_dtype(log[event_feature])
    if not (is_obj or is_bool):
        err_detected = True 
        err_msge = "Column '{}' is specified as a categorical event feature, but is not of the correct datatype. The categorical feature \
columns should be of the object dtype or of the boolean dtype.".format(event_feature)
        err_cat = 'type_error'
    return err_detected, err_msge, err_cat

#   Utility method to verify the to-be added numerical event feature 
def _verify_newNumEvFt(log, event_feature):
    ''' Validity check specifically for verifying whether the to-be added event_feature in the add_numerical_eventft(self, event_feature)
        method is valid. (Also called upon to verify the numerical event features specified upon initialization of the DynamicLogPlots class.)

        args:
            - log: pd.DataFrame 
            - event_feature: string containing the column name (in log) of the numerical event feature to-be added. 

        returns:
            - error_detected    : boolean; True if an error is detected in the to-be added numerical event feature column, False otherwise.
            - err_msge          : string; returns an adequate error message if needed. 
            - err_cat           : string = 'key_error' or 'type_error'
    '''
    err_detected, err_msge, err_cat = False, " ", " "
    cols = list(log.columns)
    if event_feature not in cols:
        err_detected = True
        err_msge = "'{}': Column with column name '{}' is not found in the given log.".format(event_feature, event_feature)
        err_cat = 'key_error'
        return err_detected, err_msge, err_cat
    if not pd.api.types.is_numeric_dtype(log[event_feature]):
        err_detected = True 
        err_msge = "Column '{}' is specified as a numerical event feature, but is not of a numeric dtype.".format(event_feature)
        err_cat = 'type_error'
    return err_detected, err_msge, err_cat
